# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
Calculating root mean square quantities --- :mod:`MDAnalysis.analysis.rms`
==========================================================================

:Author: Oliver Beckstein, David L. Dotson, John Detlefs
:Year: 2016
:Copyright: Lesser GNU Public License v2.1+

.. versionadded:: 0.7.7
.. versionchanged:: 0.11.0
   Added :class:`RMSF` analysis.
.. versionchanged:: 0.16.0
   Refactored RMSD to fit AnalysisBase API

The module contains code to analyze root mean square quantities such
as the coordinat root mean square distance (:class:`RMSD`) or the
per-residue root mean square fluctuations (:class:`RMSF`).

This module uses the fast QCP algorithm [Theobald2005]_ to calculate
the root mean square distance (RMSD) between two coordinate sets (as
implemented in
:func:`MDAnalysis.lib.qcprot.CalcRMSDRotationalMatrix`).

When using this module in published work please cite [Theobald2005]_.


See Also
--------
:mod:`MDAnalysis.analysis.align`
   aligning structures based on RMSD
:mod:`MDAnalysis.lib.qcprot`
   implements the fast RMSD algorithm.


Example applications
--------------------

Calculating RMSD for multiple domains
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

In this example we will globally fit a protein to a reference
structure and investigate the relative movements of domains by
computing the RMSD of the domains to the reference. The example is a
DIMS trajectory of adenylate kinase, which samples a large
closed-to-open transition. The protein consists of the CORE, LID, and
NMP domain.

* superimpose on the closed structure (frame 0 of the trajectory),
  using backbone atoms

* calculate the backbone RMSD and RMSD for CORE, LID, NMP (backbone atoms)

The trajectory is included with the test data files. The data in
:attr:`RMSD.results.rmsd` is plotted with :func:`matplotlib.pyplot.plot` (see Figure :ref:`RMSD plot figure <figure-RMSD>`)::

   import MDAnalysis
   from MDAnalysis.tests.datafiles import PSF,DCD,CRD
   u = MDAnalysis.Universe(PSF,DCD)
   ref = MDAnalysis.Universe(PSF,DCD)     # reference closed AdK (1AKE) (with the default ref_frame=0)
   #ref = MDAnalysis.Universe(PSF,CRD)    # reference open AdK (4AKE)

   import MDAnalysis.analysis.rms

   R = MDAnalysis.analysis.rms.RMSD(u, ref,
              select="backbone",             # superimpose on whole backbone of the whole protein
              groupselections=["backbone and (resid 1-29 or resid 60-121 or resid 160-214)",   # CORE
                               "backbone and resid 122-159",                                   # LID
                               "backbone and resid 30-59"])                                    # NMP
   R.run()

   import matplotlib.pyplot as plt
   rmsd = R.results.rmsd.T   # transpose makes it easier for plotting
   time = rmsd[1]
   fig = plt.figure(figsize=(4,4))
   ax = fig.add_subplot(111)
   ax.plot(time, rmsd[2], 'k-',  label="all")
   ax.plot(time, rmsd[3], 'k--', label="CORE")
   ax.plot(time, rmsd[4], 'r--', label="LID")
   ax.plot(time, rmsd[5], 'b--', label="NMP")
   ax.legend(loc="best")
   ax.set_xlabel("time (ps)")
   ax.set_ylabel(r"RMSD ($\\AA$)")
   fig.savefig("rmsd_all_CORE_LID_NMP_ref1AKE.pdf")

.. _figure-RMSD:

.. figure:: /images/RSMD_plot.png
      :scale: 50 %
      :alt: RMSD plot

      RMSD plot for backbone and CORE, LID, NMP domain of the protein.


Functions
---------

.. autofunction:: rmsd

Analysis classes
----------------

.. autoclass:: RMSD
   :members:
   :inherited-members:

   .. attribute:: results.rmsd

       Contains the time series of the RMSD as an N×3 :class:`numpy.ndarray`
       array with content ``[[frame, time (ps), RMSD (A)], [...], ...]``.

       .. versionadded:: 2.0.0

   .. attribute:: rmsd

       Alias to the :attr:`results.rmsd` attribute.

       .. deprecated:: 2.0.0
          Will be removed in MDAnalysis 3.0.0. Please use :attr:`results.rmsd`
          instead.


.. autoclass:: RMSF
   :members:
   :inherited-members:

   .. attribute:: results.rmsf

      Results are stored in this N-length :class:`numpy.ndarray` array,
      giving RMSFs for each of the given atoms.

      .. versionadded:: 2.0.0

   .. attribute:: rmsf

      Alias to the :attr:`results.rmsf` attribute.

      .. deprecated:: 2.0.0
         Will be removed in MDAnalysis 3.0.0. Please use :attr:`results.rmsf`
         instead.

"""
import numpy as np

import logging
import warnings

from ..lib import qcprot as qcp
from ..analysis.base import AnalysisBase, ResultsGroup
from ..exceptions import SelectionError
from ..lib.util import asiterable, iterable, get_weights


logger = logging.getLogger("MDAnalysis.analysis.rmsd")


def rmsd(a, b, weights=None, center=False, superposition=False):
    r"""Returns RMSD between two coordinate sets `a` and `b`.

    `a` and `b` are arrays of the coordinates of N atoms of shape
    :math:`N times 3` as generated by, e.g.,
    :meth:`MDAnalysis.core.groups.AtomGroup.positions`.

    Note
    ----
    If you use trajectory data from simulations performed under **periodic
    boundary conditions** then you *must make your molecules whole* before
    performing RMSD calculations so that the centers of mass of the mobile and
    reference structure are properly superimposed.


    Parameters
    ----------
    a : array_like
        coordinates to align to `b`
    b : array_like
        coordinates to align to (same shape as `a`)
    weights : array_like (optional)
        1D array with weights, use to compute weighted average
    center : bool (optional)
        subtract center of geometry before calculation. With weights given
        compute weighted average as center.
    superposition : bool (optional)
        perform a rotational and translational superposition with the fast QCP
        algorithm [Theobald2005]_ before calculating the RMSD; implies
        ``center=True``.

    Returns
    -------
    rmsd : float
        RMSD between `a` and `b`

    Notes
    -----
    The RMSD :math:`\rho(t)` as a function of time is calculated as

    .. math::

       \rho(t) = \sqrt{\frac{1}{N} \sum_{i=1}^N w_i \left(\mathbf{x}_i(t)
                                - \mathbf{x}_i^{\text{ref}}\right)^2}

    It is the Euclidean distance in configuration space of the current
    configuration (possibly after optimal translation and rotation) from a
    reference configuration divided by :math:`1/\sqrt{N}` where :math:`N` is
    the number of coordinates.

    The weights :math:`w_i` are calculated from the input weights
    `weights` :math:`w'_i` as relative to the mean:

    .. math::

       w_i = \frac{w'_i}{\langle w' \rangle}


    Example
    -------
    >>> import MDAnalysis as mda
    >>> from MDAnalysis.analysis.rms import rmsd
    >>> from MDAnalysis.tests.datafiles import PSF, DCD
    >>> u = mda.Universe(PSF, DCD)
    >>> bb = u.select_atoms('backbone')
    >>> A = bb.positions.copy()  # coordinates of first frame
    >>> _ = u.trajectory[-1]  # forward to last frame
    >>> B = bb.positions.copy()  # coordinates of last frame
    >>> rmsd(A, B, center=True)
    6.838544558398293


    .. versionchanged:: 0.8.1
       *center* keyword added
    .. versionchanged:: 0.14.0
       *superposition* keyword added

    """

    a = np.asarray(a, dtype=np.float64)
    b = np.asarray(b, dtype=np.float64)
    N = b.shape[0]
    if a.shape != b.shape:
        raise ValueError("a and b must have same shape")

    # superposition only works if structures are centered
    if center or superposition:
        # make copies (do not change the user data!)
        # weights=None is equivalent to all weights 1
        a = a - np.average(a, axis=0, weights=weights)
        b = b - np.average(b, axis=0, weights=weights)

    if weights is not None:
        if len(weights) != len(a):
            raise ValueError("weights must have same length as a and b")
        # weights are constructed as relative to the mean
        weights = np.asarray(weights, dtype=np.float64) / np.mean(weights)

    if superposition:
        return qcp.CalcRMSDRotationalMatrix(a, b, N, None, weights)
    else:
        if weights is not None:
            return np.sqrt(np.sum(weights[:, np.newaxis] * ((a - b) ** 2)) / N)
        else:
            return np.sqrt(np.sum((a - b) ** 2) / N)


def process_selection(select):
    """Return a canonical selection dictionary.

    Parameters
    ----------
    select : str or tuple or dict

        - `str` -> Any valid string selection
        - `dict` -> ``{'mobile':sel1, 'reference':sel2}``
        - `tuple` -> ``(sel1, sel2)``

    Returns
    -------
    dict
        selections for 'reference' and 'mobile'. Values are guarenteed to be
        iterable (so that one can provide selections to retain order)

    Notes
    -----
    The dictionary input for `select` can be generated by
    :func:`fasta2select` based on a ClustalW_ or STAMP_ sequence alignment.
    """

    if isinstance(select, str):
        select = {"reference": str(select), "mobile": str(select)}
    elif type(select) is tuple:
        try:
            select = {"mobile": select[0], "reference": select[1]}
        except IndexError:
            raise IndexError(
                "select must contain two selection strings "
                "(reference, mobile)"
            ) from None
    elif type(select) is dict:
        # compatability hack to use new nomenclature
        try:
            select["mobile"]
            select["reference"]
        except KeyError:
            raise KeyError(
                "select dictionary must contain entries for keys "
                "'mobile' and 'reference'."
            ) from None
    else:
        raise TypeError("'select' must be either a string, 2-tuple, or dict")
    select["mobile"] = asiterable(select["mobile"])
    select["reference"] = asiterable(select["reference"])
    return select


class RMSD(AnalysisBase):
    r"""Class to perform RMSD analysis on a trajectory.

    The RMSD will be computed for two groups of atoms and all frames in the
    trajectory belonging to `atomgroup`. The groups of atoms are obtained by
    applying the selection selection `select` to the changing `atomgroup` and
    the fixed `reference`.

    Note
    ----
    If you use trajectory data from simulations performed under **periodic
    boundary conditions** then you *must make your molecules whole* before
    performing RMSD calculations so that the centers of mass of the selected
    and reference structure are properly superimposed.


    Run the analysis with :meth:`RMSD.run`, which stores the results
    in the array :attr:`RMSD.results.rmsd`.


    .. versionchanged:: 1.0.0
       ``save()`` method was removed, use ``np.savetxt()`` on
       :attr:`RMSD.results.rmsd` instead.
    .. versionchanged:: 2.0.0
       :attr:`rmsd` results are now stored in a
       :class:`MDAnalysis.analysis.base.Results` instance.
    .. versionchanged:: 2.8.0
       introduced :meth:`get_supported_backends` allowing for parallel
       execution on ``multiprocessing`` and ``dask`` backends.
    """

    _analysis_algorithm_is_parallelizable = True

    @classmethod
    def get_supported_backends(cls):
        return (
            "serial",
            "multiprocessing",
            "dask",
        )

    def __init__(
        self,
        atomgroup,
        reference=None,
        select="all",
        groupselections=None,
        weights=None,
        weights_groupselections=False,
        tol_mass=0.1,
        ref_frame=0,
        **kwargs,
    ):
        r"""Parameters
        ----------
        atomgroup : AtomGroup or Universe
            Group of atoms for which the RMSD is calculated. If a trajectory is
            associated with the atoms then the computation iterates over the
            trajectory.
        reference : AtomGroup or Universe (optional)
            Group of reference atoms; if ``None`` then the current frame of
            `atomgroup` is used.
        select : str or dict or tuple (optional)
            The selection to operate on; can be one of:

            1. any valid selection string for
               :meth:`~MDAnalysis.core.groups.AtomGroup.select_atoms` that
               produces identical selections in `atomgroup` and `reference`; or

            2. a dictionary ``{'mobile': sel1, 'reference': sel2}`` where *sel1*
               and *sel2* are valid selection strings that are applied to
               `atomgroup` and `reference` respectively (the
               :func:`MDAnalysis.analysis.align.fasta2select` function returns such
               a dictionary based on a ClustalW_ or STAMP_ sequence alignment); or

            3. a tuple ``(sel1, sel2)``

            When using 2. or 3. with *sel1* and *sel2* then these selection strings
            are applied to `atomgroup` and `reference` respectively and should
            generate *groups of equivalent atoms*.  *sel1* and *sel2* can each also
            be a *list of selection strings* to generate a
            :class:`~MDAnalysis.core.groups.AtomGroup` with defined atom order as
            described under :ref:`ordered-selections-label`).

        groupselections : list (optional)
            A list of selections as described for `select`, with the difference
            that these selections are *always applied to the full universes*,
            i.e., ``atomgroup.universe.select_atoms(sel1)`` and
            ``reference.universe.select_atoms(sel2)``. Each selection describes
            additional RMSDs to be computed *after the structures have been
            superimposed* according to `select`. No additional fitting is
            performed.The output contains one additional column for each
            selection.

            .. Note:: Experimental feature. Only limited error checking
                      implemented.

        weights : {"mass", ``None``} or array_like (optional)
             1. "mass" will use masses as weights for both `select` and `groupselections`.

             2. ``None`` will weigh each atom equally for both `select` and `groupselections`.

             3. If 1D float array of the same length as `atomgroup` is provided,
             use each element of the `array_like` as a weight for the
             corresponding atom in `select`, and assumes ``None`` for `groupselections`.

        weights_groupselections : False or list of {"mass", ``None`` or array_like} (optional)
             1. ``False`` will apply imposed weights to `groupselections` from
             ``weights`` option if ``weights`` is either ``"mass"`` or ``None``.
             Otherwise will assume a list of length equal to length of
             `groupselections` filled with ``None`` values.

             2. A list of {"mass", ``None`` or array_like} with the length of `groupselections`
             will apply the weights to `groupselections` correspondingly.

        tol_mass : float (optional)
             Reject match if the atomic masses for matched atoms differ by more
             than `tol_mass`.
        ref_frame : int (optional)
             frame index to select frame from `reference`
        verbose : bool (optional)
             Show detailed progress of the calculation if set to ``True``; the
             default is ``False``.

        Raises
        ------
        SelectionError
             If the selections from `atomgroup` and `reference` do not match.
        TypeError
             If `weights` or `weights_groupselections` is not of the appropriate type;
             see also :func:`MDAnalysis.lib.util.get_weights`
        ValueError
             If `weights` are not compatible with `atomgroup` (not the same
             length) or if it is not a 1D array (see
             :func:`MDAnalysis.lib.util.get_weights`).

             A :exc:`ValueError` is also raised if the length of `weights_groupselections`
             are not compatible with `groupselections`.

        Notes
        -----
        The root mean square deviation :math:`\rho(t)` of a group of :math:`N`
        atoms relative to a reference structure as a function of time is
        calculated as

        .. math::

           \rho(t) = \sqrt{\frac{1}{N} \sum_{i=1}^N w_i \left(\mathbf{x}_i(t)
                                    - \mathbf{x}_i^{\text{ref}}\right)^2}

        The weights :math:`w_i` are calculated from the input weights `weights`
        :math:`w'_i` as relative to the mean of the input weights:

        .. math::

           w_i = \frac{w'_i}{\langle w' \rangle}

        The selected coordinates from `atomgroup` are optimally superimposed
        (translation and rotation) on the `reference` coordinates at each time step
        as to minimize the RMSD. Douglas Theobald's fast QCP algorithm
        [Theobald2005]_ is used for the rotational superposition and to calculate
        the RMSD (see :mod:`MDAnalysis.lib.qcprot` for implementation details).

        The class runs various checks on the input to ensure that the two atom
        groups can be compared. This includes a comparison of atom masses (i.e.,
        only the positions of atoms of the same mass will be considered to be
        correct for comparison). If masses should not be checked, just set
        `tol_mass` to a large value such as 1000.

        .. _ClustalW: http://www.clustal.org/
        .. _STAMP: http://www.compbio.dundee.ac.uk/manuals/stamp.4.2/


        See Also
        --------
        rmsd


        .. versionadded:: 0.7.7
        .. versionchanged:: 0.8
           `groupselections` added
        .. versionchanged:: 0.16.0
           Flexible weighting scheme with new `weights` keyword.
        .. deprecated:: 0.16.0
           Instead of ``mass_weighted=True`` (removal in 0.17.0) use new
           ``weights='mass'``; refactored to fit with AnalysisBase API
        .. versionchanged:: 0.17.0
           removed deprecated `mass_weighted` keyword; `groupselections`
           are *not* rotationally superimposed any more.
        .. versionchanged:: 1.0.0
           `filename` keyword was removed.

        """
        super(RMSD, self).__init__(atomgroup.universe.trajectory, **kwargs)
        self.atomgroup = atomgroup
        self.reference = reference if reference is not None else self.atomgroup

        select = process_selection(select)
        self.groupselections = (
            [process_selection(s) for s in groupselections]
            if groupselections is not None
            else []
        )
        self.weights = weights
        self.tol_mass = tol_mass
        self.ref_frame = ref_frame
        self.weights_groupselections = weights_groupselections
        self.ref_atoms = self.reference.select_atoms(*select["reference"])
        self.mobile_atoms = self.atomgroup.select_atoms(*select["mobile"])

        if len(self.ref_atoms) != len(self.mobile_atoms):
            err = (
                "Reference and trajectory atom selections do "
                "not contain the same number of atoms: "
                "N_ref={0:d}, N_traj={1:d}".format(
                    self.ref_atoms.n_atoms, self.mobile_atoms.n_atoms
                )
            )
            logger.exception(err)
            raise SelectionError(err)
        logger.info(
            "RMS calculation " "for {0:d} atoms.".format(len(self.ref_atoms))
        )
        mass_mismatches = (
            np.absolute((self.ref_atoms.masses - self.mobile_atoms.masses))
            > self.tol_mass
        )

        if np.any(mass_mismatches):
            # diagnostic output:
            logger.error("Atoms: reference | mobile")
            for ar, at in zip(self.ref_atoms, self.mobile_atoms):
                if ar.name != at.name:
                    logger.error(
                        "{0!s:>4} {1:3d} {2!s:>3} {3!s:>3} {4:6.3f}"
                        "|  {5!s:>4} {6:3d} {7!s:>3} {8!s:>3}"
                        "{9:6.3f}".format(
                            ar.segid,
                            ar.resid,
                            ar.resname,
                            ar.name,
                            ar.mass,
                            at.segid,
                            at.resid,
                            at.resname,
                            at.name,
                            at.mass,
                        )
                    )
            errmsg = (
                "Inconsistent selections, masses differ by more than"
                "{0:f}; mis-matching atoms"
                "are shown above.".format(self.tol_mass)
            )
            logger.error(errmsg)
            raise SelectionError(errmsg)
        del mass_mismatches

        # TODO:
        # - make a group comparison a class that contains the checks above
        # - use this class for the *select* group and the additional
        #   *groupselections* groups each a dict with reference/mobile
        self._groupselections_atoms = [
            {
                "reference": self.reference.universe.select_atoms(
                    *s["reference"]
                ),
                "mobile": self.atomgroup.universe.select_atoms(*s["mobile"]),
            }
            for s in self.groupselections
        ]
        # sanity check
        for igroup, (sel, atoms) in enumerate(
            zip(self.groupselections, self._groupselections_atoms)
        ):
            if len(atoms["mobile"]) != len(atoms["reference"]):
                logger.exception("SelectionError: Group Selection")
                raise SelectionError(
                    "Group selection {0}: {1} | {2}: Reference and trajectory "
                    "atom selections do not contain the same number of atoms: "
                    "N_ref={3}, N_traj={4}".format(
                        igroup,
                        sel["reference"],
                        sel["mobile"],
                        len(atoms["reference"]),
                        len(atoms["mobile"]),
                    )
                )

        # check weights type
        acceptable_dtypes = (np.dtype("float64"), np.dtype("int64"))
        msg = (
            "weights should only be 'mass', None or 1D float array."
            "For weights on groupselections, "
            "use **weight_groupselections**"
        )

        if iterable(self.weights):
            element_lens = []
            for element in self.weights:
                if iterable(element):
                    element_lens.append(len(element))
                else:
                    element_lens.append(1)
            if np.unique(element_lens).size > 1:
                # jagged data structure
                raise TypeError(msg)
            if np.array(element).dtype not in acceptable_dtypes:
                raise TypeError(msg)

        if iterable(self.weights) or self.weights != "mass":
            get_weights(self.mobile_atoms, self.weights)

        if self.weights_groupselections:
            if len(self.weights_groupselections) != len(self.groupselections):
                raise ValueError(
                    "Length of weights_groupselections is not equal to "
                    "length of groupselections "
                )
            for weights, atoms, selection in zip(
                self.weights_groupselections,
                self._groupselections_atoms,
                self.groupselections,
            ):
                try:
                    if iterable(weights) or weights != "mass":
                        get_weights(atoms["mobile"], weights)
                except Exception as e:
                    raise type(e)(
                        str(e)
                        + " happens in selection %s" % selection["mobile"]
                    )

    def _prepare(self):
        self._n_atoms = self.mobile_atoms.n_atoms
        if not self.weights_groupselections:
            if not iterable(
                self.weights
            ):  # apply 'mass' or 'None' to groupselections
                self.weights_groupselections = [self.weights] * len(
                    self.groupselections
                )
            else:
                self.weights_groupselections = [None] * len(
                    self.groupselections
                )

        for igroup, (weights, atoms) in enumerate(
            zip(self.weights_groupselections, self._groupselections_atoms)
        ):
            if str(weights) == "mass":
                self.weights_groupselections[igroup] = atoms["mobile"].masses
            if weights is not None:
                self.weights_groupselections[igroup] = np.asarray(
                    self.weights_groupselections[igroup], dtype=np.float64
                ) / np.mean(self.weights_groupselections[igroup])
        # add the array of weights to weights_select
        self.weights_select = get_weights(self.mobile_atoms, self.weights)
        self.weights_ref = get_weights(self.ref_atoms, self.weights)
        if self.weights_select is not None:
            self.weights_select = np.asarray(
                self.weights_select, dtype=np.float64
            ) / np.mean(self.weights_select)
            self.weights_ref = np.asarray(
                self.weights_ref, dtype=np.float64
            ) / np.mean(self.weights_ref)

        current_frame = self.reference.universe.trajectory.ts.frame

        try:
            # Move to the ref_frame
            # (coordinates MUST be stored in case the ref traj is advanced
            # elsewhere or if ref == mobile universe)
            self.reference.universe.trajectory[self.ref_frame]
            self._ref_com = self.ref_atoms.center(self.weights_ref)
            # makes a copy
            self._ref_coordinates = self.ref_atoms.positions - self._ref_com
            if self._groupselections_atoms:
                self._groupselections_ref_coords64 = [
                    (
                        self.reference.select_atoms(
                            *s["reference"]
                        ).positions.astype(np.float64)
                    )
                    for s in self.groupselections
                ]
        finally:
            # Move back to the original frame
            self.reference.universe.trajectory[current_frame]

        self._ref_coordinates64 = self._ref_coordinates.astype(np.float64)

        if self._groupselections_atoms:
            # Only carry out a rotation if we want to calculate secondary
            # RMSDs.
            # R: rotation matrix that aligns r-r_com, x~-x~com
            #    (x~: selected coordinates, x: all coordinates)
            # Final transformed traj coordinates: x' = (x-x~_com)*R + ref_com
            self._rot = np.zeros(9, dtype=np.float64)  # allocate space
            self._R = self._rot.reshape(3, 3)
        else:
            self._rot = None

        self.results.rmsd = np.zeros(
            (self.n_frames, 3 + len(self._groupselections_atoms))
        )

        self._mobile_coordinates64 = self.mobile_atoms.positions.copy().astype(
            np.float64
        )

    def _get_aggregator(self):
        return ResultsGroup(lookup={"rmsd": ResultsGroup.ndarray_vstack})

    def _single_frame(self):
        mobile_com = self.mobile_atoms.center(self.weights_select).astype(
            np.float64
        )
        self._mobile_coordinates64[:] = self.mobile_atoms.positions
        self._mobile_coordinates64 -= mobile_com

        self.results.rmsd[self._frame_index, :2] = (
            self._ts.frame,
            self._trajectory.time,
        )

        if self._groupselections_atoms:
            # superimpose structures: MDAnalysis qcprot needs Nx3 coordinate
            # array with float64 datatype (float32 leads to errors up to 1e-3 in
            # RMSD). Note that R is defined in such a way that it acts **to the
            # left** so that we can easily use broadcasting and save one
            # expensive numpy transposition.

            self.results.rmsd[self._frame_index, 2] = (
                qcp.CalcRMSDRotationalMatrix(
                    self._ref_coordinates64,
                    self._mobile_coordinates64,
                    self._n_atoms,
                    self._rot,
                    self.weights_select,
                )
            )

            self._R[:, :] = self._rot.reshape(3, 3)
            # Transform each atom in the trajectory (use inplace ops to
            # avoid copying arrays) (Marginally (~3%) faster than
            # "ts.positions[:] = (ts.positions - x_com) * R + ref_com".)
            self._ts.positions[:] -= mobile_com

            # R acts to the left & is broadcasted N times.
            self._ts.positions[:] = np.dot(self._ts.positions, self._R)
            self._ts.positions += self._ref_com

            # 2) calculate secondary RMSDs (without any further
            #    superposition)
            for igroup, (refpos, atoms) in enumerate(
                zip(
                    self._groupselections_ref_coords64,
                    self._groupselections_atoms,
                ),
                3,
            ):
                self.results.rmsd[self._frame_index, igroup] = rmsd(
                    refpos,
                    atoms["mobile"].positions,
                    weights=self.weights_groupselections[igroup - 3],
                    center=False,
                    superposition=False,
                )
        else:
            # only calculate RMSD by setting the Rmatrix to None (no need
            # to carry out the rotation as we already get the optimum RMSD)
            self.results.rmsd[self._frame_index, 2] = (
                qcp.CalcRMSDRotationalMatrix(
                    self._ref_coordinates64,
                    self._mobile_coordinates64,
                    self._n_atoms,
                    None,
                    self.weights_select,
                )
            )

    @property
    def rmsd(self):
        wmsg = (
            "The `rmsd` attribute was deprecated in MDAnalysis 2.0.0 and "
            "will be removed in MDAnalysis 3.0.0. Please use "
            "`results.rmsd` instead."
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.rmsd


class RMSF(AnalysisBase):
    r"""Calculate RMSF of given atoms across a trajectory.

    Note
    ----
    No RMSD-superposition is performed; it is assumed that the user is
    providing a trajectory where the protein of interest has been structurally
    aligned to a reference structure (see the Examples section below). The
    protein also has be whole because periodic boundaries are not taken into
    account.


    Run the analysis with :meth:`RMSF.run`, which stores the results
    in the array :attr:`RMSF.results.rmsf`.

    """

    @classmethod
    def get_supported_backends(cls):
        return ("serial",)

    def __init__(self, atomgroup, **kwargs):
        r"""Parameters
        ----------
        atomgroup : AtomGroup
            Atoms for which RMSF is calculated
        verbose : bool (optional)
             Show detailed progress of the calculation if set to ``True``; the
             default is ``False``.

        Raises
        ------
        ValueError
             raised if negative values are calculated, which indicates that a
             numerical overflow or underflow occured


        Notes
        -----
        The root mean square fluctuation of an atom :math:`i` is computed as the
        time average

        .. math::

          \rho_i = \sqrt{\left\langle (\mathbf{x}_i - \langle\mathbf{x}_i\rangle)^2 \right\rangle}

        No mass weighting is performed.

        This method implements an algorithm for computing sums of squares while
        avoiding overflows and underflows :footcite:p:`Welford1962`.


        Examples
        --------

        In this example we calculate the residue RMSF fluctuations by analyzing
        the :math:`\text{C}_\alpha` atoms. First we need to fit the trajectory
        to the average structure as a reference. That requires calculating the
        average structure first. Because we need to analyze and manipulate the
        same trajectory multiple times, we are going to load it into memory
        using the :mod:`~MDAnalysis.coordinates.MemoryReader`. (If your
        trajectory does not fit into memory, you will need to :ref:`write out
        intermediate trajectories <writing-trajectories>` to disk or
        :ref:`generate an in-memory universe
        <creating-in-memory-trajectory-label>` that only contains, say, the
        protein)::

           import MDAnalysis as mda
           from MDAnalysis.analysis import align

           from MDAnalysis.tests.datafiles import TPR, XTC

           u = mda.Universe(TPR, XTC, in_memory=True)
           protein = u.select_atoms("protein")

           # 1) the current trajectory contains a protein split across
           #    periodic boundaries, so we first make the protein whole and
           #    center it in the box using on-the-fly transformations
           import MDAnalysis.transformations as trans

           not_protein = u.select_atoms('not protein')
           transforms = [trans.unwrap(protein),
                         trans.center_in_box(protein, wrap=True),
                         trans.wrap(not_protein)]
           u.trajectory.add_transformations(*transforms)

           # 2) fit to the initial frame to get a better average structure
           #    (the trajectory is changed in memory)
           prealigner = align.AlignTraj(u, u, select="protein and name CA",
                                        in_memory=True).run()

           # 3) reference = average structure
           ref_coordinates = u.trajectory.timeseries(asel=protein).mean(axis=1)
           # make a reference structure (need to reshape into a 1-frame
           # "trajectory")
           reference = mda.Merge(protein).load_new(ref_coordinates[:, None, :],
                                                   order="afc")

        We created a new universe ``reference`` that contains a single frame
        with the averaged coordinates of the protein.  Now we need to fit the
        whole trajectory to the reference by minimizing the RMSD. We use
        :class:`MDAnalysis.analysis.align.AlignTraj`::

           aligner = align.AlignTraj(u, reference,
                                     select="protein and name CA",
                                     in_memory=True).run()

        The trajectory is now fitted to the reference (the RMSD is stored as
        `aligner.results.rmsd` for further inspection). Now we can calculate
        the RMSF::

           from MDAnalysis.analysis.rms import RMSF

           calphas = protein.select_atoms("name CA")
           rmsfer = RMSF(calphas, verbose=True).run()

        and plot::

           import matplotlib.pyplot as plt

           plt.plot(calphas.resnums, rmsfer.results.rmsf)



        References
        ----------
        .. footbibliography::


        .. versionadded:: 0.11.0
        .. versionchanged:: 0.16.0
           refactored to fit with AnalysisBase API
        .. deprecated:: 0.16.0
           the keyword argument `quiet` is deprecated in favor of `verbose`.
        .. versionchanged:: 0.17.0
           removed unused keyword `weights`
        .. versionchanged:: 1.0.0
           Support for the ``start``, ``stop``, and ``step`` keywords has been
           removed. These should instead be passed to :meth:`RMSF.run`.

        """
        super(RMSF, self).__init__(atomgroup.universe.trajectory, **kwargs)
        self.atomgroup = atomgroup

    def _prepare(self):
        self.sumsquares = np.zeros((self.atomgroup.n_atoms, 3))
        self.mean = self.sumsquares.copy()

    def _single_frame(self):
        k = self._frame_index
        self.sumsquares += (k / (k + 1.0)) * (
            self.atomgroup.positions - self.mean
        ) ** 2
        self.mean = (k * self.mean + self.atomgroup.positions) / (k + 1)

    def _conclude(self):
        k = self._frame_index
        self.results.rmsf = np.sqrt(self.sumsquares.sum(axis=1) / (k + 1))

        if not (self.results.rmsf >= 0).all():
            raise ValueError(
                "Some RMSF values negative; overflow "
                + "or underflow occurred"
            )

    @property
    def rmsf(self):
        wmsg = (
            "The `rmsf` attribute was deprecated in MDAnalysis 2.0.0 and "
            "will be removed in MDAnalysis 3.0.0. Please use "
            "`results.rmsd` instead."
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.rmsf



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
import numpy as np
from numpy.testing import assert_equal

import MDAnalysis as mda
from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import PSF_NAMD_GBIS, PRM

pmd = pytest.importorskip("parmed")


class BaseTestParmedParser(ParserBase):
    parser = mda.converters.ParmEdParser.ParmEdParser
    expected_attrs = [
        "ids",
        "names",
        "types",
        "masses",
        "charges",
        "altLocs",
        "occupancies",
        "tempfactors",
        "gbscreens",
        "solventradii",
        "nbindices",
        "rmins",
        "epsilons",
        "rmin14s",
        "epsilon14s",
        "elements",
        "chainIDs",
        "resids",
        "resnames",
        "resnums",
        "segids",
        "bonds",
        "ureybradleys",
        "angles",
        "dihedrals",
        "impropers",
        "cmaps",
    ]

    expected_n_atoms = 0
    expected_n_residues = 1
    expected_n_segments = 1
    expected_n_bonds = 0
    expected_n_angles = 0
    expected_n_dihedrals = 0
    expected_n_impropers = 0
    expected_n_cmaps = 0
    expected_n_ureybradleys = 0

    @pytest.fixture
    def filename(self):
        return pmd.load_file(self.ref_filename)

    @pytest.fixture
    def universe(self, filename):
        return mda.Universe(filename)

    def test_creates_universe(self, filename):
        u = mda.Universe(filename)
        assert isinstance(u, mda.Universe)

    def test_bonds_total_counts(self, top, filename):
        unique = set([(a.atom1.idx, a.atom2.idx) for a in filename.bonds])
        assert len(top.bonds.values) == len(unique)

    def test_angles_total_counts(self, top, filename):
        unique = set(
            [(a.atom1.idx, a.atom2.idx, a.atom3.idx) for a in filename.angles]
        )
        assert len(top.angles.values) == len(unique)

    def test_dihedrals_total_counts(self, top, filename):
        unique = set(
            [
                (a.atom1.idx, a.atom2.idx, a.atom3.idx, a.atom4.idx)
                for a in filename.dihedrals
            ]
        )
        assert len(top.dihedrals.values) == len(unique)

    def test_impropers_total_counts(self, top, filename):
        unique = set(
            [
                (a.atom1.idx, a.atom2.idx, a.atom3.idx, a.atom4.idx)
                for a in filename.impropers
            ]
        )
        assert len(top.impropers.values) == len(unique)

    def test_cmaps_total_counts(self, top, filename):
        unique = set(
            [
                (
                    a.atom1.idx,
                    a.atom2.idx,
                    a.atom3.idx,
                    a.atom4.idx,
                    a.atom5.idx,
                )
                for a in filename.cmaps
            ]
        )
        assert len(top.cmaps.values) == len(unique)

    def test_ureybradleys_total_counts(self, top, filename):
        unique = set(
            [(a.atom1.idx, a.atom2.idx) for a in filename.urey_bradleys]
        )
        assert len(top.ureybradleys.values) == len(unique)

    def test_elements(self, top):
        for erange, evals in zip(self.elems_ranges, self.expected_elems):
            assert_equal(
                top.elements.values[erange[0] : erange[1]],
                evals,
                "unexpected element match",
            )


class TestParmedParserPSF(BaseTestParmedParser):
    """
    PSF with CMAPs
    """

    ref_filename = PSF_NAMD_GBIS

    expected_n_atoms = 3341
    expected_n_residues = 214
    expected_n_bonds = 3365
    expected_n_angles = 6123
    expected_n_dihedrals = 8921
    expected_n_impropers = 541
    expected_n_cmaps = 212
    elems_ranges = ((100, 120),)
    # No atomic numbers set by parmed == no elements
    expected_elems = (np.array(list("NHCHCHHCHCHHHCHHHCON"), dtype=object),)

    def test_bonds_atom_counts(self, universe):
        assert len(universe.atoms[[0]].bonds) == 4
        assert len(universe.atoms[[42]].bonds) == 1

    @pytest.mark.parametrize(
        "value",
        (
            (0, 1),
            (0, 2),
            (0, 3),
            (0, 4),
        ),
    )
    def test_bonds_identity(self, top, value):
        vals = top.bonds.values
        assert value in vals or value[::-1] in vals

    def test_bond_types(self, universe):
        b1 = universe.bonds[0]

        assert b1.order == 1.0
        assert tuple(b1.indices) == (0, 1)
        assert b1.type.type is None

    def test_angles_atom_counts(self, universe):
        assert len(universe.atoms[[0]].angles), 9
        assert len(universe.atoms[[42]].angles), 2

    @pytest.mark.parametrize(
        "value",
        (
            (1, 0, 2),
            (1, 0, 3),
            (1, 0, 4),
        ),
    )
    def test_angles_identity(self, top, value):
        vals = top.angles.values
        assert value in vals or value[::-1] in vals

    def test_dihedrals_atom_counts(self, universe):
        assert len(universe.atoms[[0]].dihedrals) == 14

    @pytest.mark.parametrize(
        "value",
        (
            (0, 4, 6, 7),
            (0, 4, 6, 8),
            (0, 4, 6, 9),
            (0, 4, 17, 18),
        ),
    )
    def test_dihedrals_identity(self, top, value):
        vals = top.dihedrals.values
        assert value in vals or value[::-1] in vals

    @pytest.mark.parametrize(
        "value", ((17, 19, 21, 41, 43), (60, 62, 64, 79, 81))
    )
    def test_cmaps_identity(self, top, value):
        vals = top.cmaps.values
        assert value in vals or value[::-1] in vals


class TestParmedParserPRM(BaseTestParmedParser):
    """
    PRM
    """

    ref_filename = PRM

    expected_n_atoms = 252
    expected_n_residues = 14
    elems_ranges = ((0, 8), (30, 38))
    expected_elems = (
        np.array(["N", "H", "H", "H", "C", "H", "C", "H"], dtype=object),
        np.array(["H", "C", "H", "H", "C", "C", "H", "C"], dtype=object),
    )

    def test_bonds_atom_counts(self, universe):
        assert len(universe.atoms[[0]].bonds) == 4
        assert len(universe.atoms[[42]].bonds) == 1

    @pytest.mark.parametrize(
        "value",
        (
            (10, 11),
            (10, 12),
            (4, 6),
            (4, 10),
        ),
    )
    def test_bonds_identity(self, top, value):
        vals = top.bonds.values
        assert value in vals or value[::-1] in vals

    def test_bond_types(self, universe):
        b1 = universe.bonds[0]

        assert b1.order == 1.0
        assert tuple(b1.indices) == (0, 1)
        assert b1.type.type.k == 434
        assert b1.type.type.req == 1.010

    def test_angles_atom_counts(self, universe):
        assert len(universe.atoms[[0]].angles), 9
        assert len(universe.atoms[[42]].angles), 2

    @pytest.mark.parametrize(
        "value",
        (
            (11, 10, 12),
            (10, 12, 14),
            (6, 4, 10),
            (4, 10, 11),
        ),
    )
    def test_angles_identity(self, top, value):
        vals = top.angles.values
        assert value in vals or value[::-1] in vals

    def test_dihedrals_atom_counts(self, universe):
        assert len(universe.atoms[[0]].dihedrals) == 14

    @pytest.mark.parametrize(
        "value",
        (
            (11, 10, 12, 14),
            (10, 12, 14, 16),
        ),
    )
    def test_dihedrals_identity(self, top, value):
        vals = top.dihedrals.values
        assert value in vals or value[::-1] in vals

    def test_dihedral_types(self, universe):
        ag = universe.atoms[[10, 12, 14, 16]]
        dih = universe.dihedrals.atomgroup_intersection(ag, strict=True)[0]
        assert len(dih.type) == 4
        for i, (phi_k, per) in enumerate(
            (
                (2.0, 1),
                (2.0, 2),
                (0.4, 3),
                (0.0, 4),
            )
        ):
            assert dih.type[i].type.phi_k == phi_k
            assert dih.type[i].type.per == per


def test_old_import_warning():
    wmsg = "Please import the ParmEd classes from MDAnalysis.converters"
    with pytest.warns(DeprecationWarning, match=wmsg):
        from MDAnalysis.topology.ParmEdParser import squash_identical



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
Extended PDB topology parser
============================

.. versionadded:: 0.8

This topology parser uses a PDB file to build a minimum internal structure
representation (list of atoms). The only difference from
:mod:`~MDAnalysis.topology.PDBParser` is that this parser reads a
non-standard PDB-like format in which residue numbers can be five digits
instead of four.

The topology reader reads a PDB file line by line and ignores atom numbers but
reads residue numbers up to 99,999 correctly. If you have systems containing at
least 100,000 residues then you need to use a different file format that can
handle such residue numbers.

See Also
--------
* :mod:`MDAnalysis.topology.PDBParser`
* :class:`MDAnalysis.coordinates.PDB.ExtendedPDBReader`
* :class:`MDAnalysis.core.universe.Universe`


Classes
-------

.. autoclass:: ExtendedPDBParser
   :members:
   :inherited-members:

"""

from . import PDBParser


class ExtendedPDBParser(PDBParser.PDBParser):
    """Parser that handles non-standard "extended" PDB file.

     Extended PDB files (MDAnalysis format specifier *XPDB*) may contain residue
     sequence numbers up to 99,999 by utilizing the insertion character field of
     the PDB standard.

     Creates a Topology with the following Attributes (if present):
      - serials
      - names
      - altLocs
      - chainids
      - tempfactors
      - occupancies
      - resids
      - resnames
      - segids
      - elements
      - bonds
      - formalcharges

    .. note::

       By default, atomtypes and masses will be guessed on Universe creation.
       This may change in release 3.0.
       See :ref:`Guessers` for more information.


     See Also
     --------
     :class:`MDAnalysis.coordinates.PDB.ExtendedPDBReader`

     .. versionadded:: 0.8
    """

    format = "XPDB"



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
Topology readers --- :mod:`MDAnalysis.topology`
===============================================

This submodule contains the topology readers. A topology file supplies the list
of atoms in the system, their connectivity and possibly additional information
such as B-factors, partial charges, etc. The details depend on the file format
and not every topology file provides all (or even any) additional data. This
data is made accessible through AtomGroup properties.

As a minimum, all topology parsers will provide atom ids, atom types, masses,
resids, resnums and segids as well as assigning all atoms to residues and all
residues to segments.  For systems without residues and segments, this results
in there being a single residue and segment to which all atoms belong.
In the event that this happens, a UserWarning will always be issued.

The following table lists the currently supported topology formats along with
the attributes they provide.

.. _`Supported topology formats`:

.. table:: Table of Supported Topology Formats

   ================= ========== ================= ===================================================
   Name              extension  attributes        remarks
   ================= ========== ================= ===================================================
   CHARMM/XPLOR PSF  psf        resnames,         CHARMM_/XPLOR/NAMD_ topology format;
                                names, types,     :mod:`MDAnalysis.topology.PSFParser`
                                charges,
                                bonds, angles,
                                dihedrals,
                                impropers

   CHARMM CARD [#a]_ crd        names,            "CARD" coordinate output from CHARMM_; deals with
                                tempfactors,      either standard or EXTended format;
                                resnames,         :mod:`MDAnalysis.topology.CRDParser`

   Brookhaven [#a]_  pdb/ent    names, bonds,     a simplified PDB format (as used in MD simulations)
                                resids, resnums,  is read by default
                                types,
                                chainids,
                                occupancies,
                                tempfactors,
                                resids, icodes,
                                resnames,
                                segids,

   XPDB [#a]_        pdb        As PDB except     Extended PDB format as used by e.g., NAMD_
                                icodes            (can use 5-digit residue numbers). To use, specify
                                                  the format "XPBD" explicitly:
                                                  ``Universe(..., topology_format="XPDB")``.
                                                  Module :mod:`MDAnalysis.coordinates.PDB`

   PQR [#a]_         pqr        names, charges,   PDB-like but whitespace-separated files with charge
                                types,            and radius information as used by, e.g., APBS_.
                                radii, resids,    :mod:`MDAnalysis.topology.PQRParser`
                                resnames, icodes,
                                segids

   PDBQT [#a]_       pdbqt      names, types,     file format used by AutoDock_ with atom types and
                                altLocs, charges, partial charges. Module:
                                resnames,         :mod:`MDAnalysis.topology.PDBQTParser`
                                resids,
                                icodes,
                                occupancies,
                                tempfactors,
                                segids,

   GROMOS96 [#a]_    gro        names, resids,    GROMOS96 coordinate file (used e.g., by Gromacs_)
                                resnames,         :mod:`MDAnalysis.topology.GROParser`

   Amber             top,       names, charges    simple Amber_ format reader (only supports a subset
                     prmtop,    type_indices,     of flags);
                     parm7      types,            :mod:`MDAnalysis.topology.TOPParser`
                                resnames,

   DESRES [#a]_      dms        names, numbers,   DESRES molecular structure reader (only supports
                                masses, charges,  the atom and bond records) as used by Desmond_ and Anton;
                                chainids, resids, :mod:`MDAnalysis.topology.DMSParser`
                                resnames, segids,
                                radii,

   TPR [#b]_         tpr        names, types,     Gromacs_ portable run input reader (limited
                                resids, resnames, experimental support for some of the more recent
                                charges, bonds,   versions of the file format);
                                masses, moltypes, :mod:`MDAnalysis.topology.TPRParser`
                                molnums

   ITP               itp        names, types,     Gromacs_ include topology file;
                                resids, resnames, :mod:`MDAnalysis.topology.ITPParser`
                                charges, bonds,
                                masses, segids,
                                moltypes,
                                chargegroups

   MOL2 [#a]_        mol2       ids, names,       Tripos MOL2 molecular structure format;
                                types, resids,    :mod:`MDAnalysis.topology.MOL2Parser`
                                charges, bonds,
                                resnames,

   LAMMPS [#a]_      data       ids, types,       LAMMPS_ Data file parser
                                masses, charges,  :mod:`MDAnalysis.topology.LAMMPSParser`
                                resids, bonds,
                                angles, dihedrals

   LAMMPS [#a]_      lammpsdump id, masses        LAMMPS_ ascii dump file reader
                                                  :mod:`MDAnalysis.topology.LAMMPSParser`

   XYZ [#a]_         xyz        names             XYZ File Parser.  Reads only the labels from atoms
                                                  and constructs minimal topology data.
                                                  :mod:`MDAnalysis.topology.XYZParser`

   TXYZ [#a]_        txyz,      names, atomids,   Tinker_ XYZ File Parser. Reads atom labels, numbers
                     arc        masses, types,    and connectivity.
                                bonds             :mod:`MDAnalysis.topology.TXYZParser`

   GAMESS [#a]_      gms,       names,            GAMESS_ output parser. Read only atoms of assembly
                     log        atomic charges,   section (atom, elems and coords) and construct
                                                  topology.
                                                  :mod:`MDAnalysis.topology.GMSParser`

   DL_POLY [#a]_     config,    ids, names        `DL_POLY`_ CONFIG or HISTORY file.  Reads only the
                     history                      atom names. If atoms are written out of order, will
                                                  correct the order.
                                                  :mod:`MDAnalysis.topology.DLPolyParser`

   Hoomd XML         xml        types, charges,   `HOOMD XML`_ topology file.  Reads atom types,
                                radii, masses     masses, and charges if possible. Also reads bonds,
                                bonds, angles,    angles, and dihedrals.
                                dihedrals         :mod:`MDAnalysis.topology.HoomdXMLParser`

   GSD [#a]_         gsd        types, charges,   HOOMD_ GSD topology file.  Reads atom types,
                                radii, masses     masses, and charges if possible. Also reads bonds,
                                bonds, angles,    angles, and dihedrals.
                                dihedrals         :mod:`MDAnalysis.topology.GSDParser`

   MMTF [#a]_        mmtf       altLocs,          `Macromolecular Transmission Format (MMTF)`_. An
                                tempfactors,      efficient compact format for biomolecular
                                charges, masses,  structures.
                                names, bonds,
                                occupancies,
                                types, icodes,
                                resnames, resids,
                                segids, models

   FHIAIMS [#a]_     in         names             `FHI-AIMS`_ File Parser.  Reads only the labels from
                                                  atoms and constructs minimal topology data.
                                                  :mod:`MDAnalysis.topology.FHIAIMSParser`
   ================= ========== ================= ===================================================

.. [#a] This format can also be used to provide *coordinates* so that
        it is possible to create a full
        :mod:`~MDAnalysis.core.universe.Universe` by simply providing
        a file of this format as the sole argument to
        :mod:`~MDAnalysis.core.universe.Universe`: ``u =
        Universe(filename)``

.. [#b] The Gromacs TPR format contains coordinate information but
        parsing coordinates from a TPR file is currently not implemented
        in :mod:`~MDAnalysis.topology.TPRParser`.

Note
----
:ref:`Coordinates` with the :ref:`Supported coordinate formats`


.. _CHARMM: https://www.charmm.org/charmm/
.. _HOOMD XML: http://codeblue.umich.edu/hoomd-blue/doc/page_xml_file_format.html

.. Note::

   The HOOMD XML file format has been deprecated in favor of the GSD.
   https://hoomd-blue.readthedocs.io/en/v2.9.3/module-deprecated-dump.html
   Transitioning to the GSD file format is advisable for working with HOOMD-blue simulations.

.. _HOOMD: http://glotzerlab.engin.umich.edu/hoomd-blue/
.. _NAMD: http://www.ks.uiuc.edu/Research/namd/
.. _LAMMPS: https://lammps.sandia.gov/
.. _Gromacs: http://www.gromacs.org/
.. _Amber: http://ambermd.org/
.. _Desmond: https://www.deshawresearch.com/resources_desmond.html
.. _Tinker: https://dasher.wustl.edu/tinker/
.. _DL_POLY: https://www.scd.stfc.ac.uk/Pages/DL_POLY.aspx
.. _AutoDock: http://autodock.scripps.edu/
.. _APBS: https://apbs.readthedocs.io/en/latest/
.. _Macromolecular Transmission Format (MMTF): https://www.rcsb.org/news/feature/65a1af31c76ca3abcc925d0c
.. _FHI-AIMS: https://aimsclub.fhi-berlin.mpg.de/
.. _GAMESS: https://www.msg.chem.iastate.edu/gamess/


.. _topology-parsers-developer-notes:


Developer Notes
---------------

.. versionadded:: 0.8
.. versionchanged:: 0.16.0
   The new array-based topology system completely replaced the old
   system that was based on a list of Atom instances.
.. versionchanged:: 2.0.0
   The ParmEdParser was moved to the :mod:`~MDAnalysis.converters` module

Topology information consists of data that do not change over time,
i.e. information that is the same for all time steps of a
trajectory. This includes

* identity of atoms (name, type, number, partial charge, ...) and to
  which residue and segment they belong; atoms are identified in
  MDAnalysis by their :attr:`~MDAnalysis.core.groups.Atom.index`,
  an integer number starting at 0 and incremented in the order of
  atoms found in the topology.

* bonds (pairs of atoms)

* angles (triplets of atoms)

* dihedral angles (quadruplets of atoms) — proper and improper
  dihedrals should be treated separately

Topology readers are generally called "parsers" in MDAnalysis (for
historical reasons and in order to distinguish them from coordinate
"readers"). All parsers are derived from
:class:`MDAnalysis.topology.base.TopologyReaderBase` and have a
:meth:`~MDAnalysis.topology.base.TopologyReaderBase.parse` method that
returns a :class:`MDAnalysis.core.topology.Topology` instance.


atoms
~~~~~~

The **atoms** appear to the user as an array of
:class:`~MDAnalysis.core.groups.Atom` instances. However, under the
hood this is essentially only an array of atom indices that are used
to index the various components of the topology database
:class:`~MDAnalysis.core.topology.Topology`. The parser needs to
initialize the :class:`~MDAnalysis.core.topology.Topology` with the
data read from the topology file.


See Also
--------
:ref:`topology-system-label`


bonds
~~~~~~

**Bonds** are represented as a :class:`tuple` of :class:`tuple`. Each tuple
contains two atom numbers, which indicate the atoms between which the
bond is formed. Only one of the two permutations is stored, typically
the one with the lower atom number first.


bondorder
~~~~~~~~~~

Some **bonds** have additional information called **order**. When available
this is stored in a dictionary of format {bondtuple:order}.  This extra
information is then passed to Bond initialisation in u._init_bonds()


angles
~~~~~~~

**Angles** are represented by a :class:`list` of :class:`tuple`. Each
tuple contains three atom numbers.  The second of these numbers
represents the apex of the angle.


dihedrals
~~~~~~~~~

**Proper dihedral angles** are represented by a :class:`list` of :class:`tuple`. Each
tuple contains four atom numbers. The angle of the torsion
is defined by the angle between the planes formed by atoms 1, 2, and 3,
and 2, 3, and 4.


impropers
~~~~~~~~~

**Improper dihedral angles** are represented by a :class:`list` of :class:`tuple`. Each
tuple contains four atom numbers.  The angle of the improper torsion
is again defined by the angle between the planes formed by atoms 1, 2, and 3,
and 2, 3, and 4.  Improper dihedrals differ from regular dihedrals as the
four atoms need not be sequentially bonded, and are instead often all bonded
to the second atom.

"""

__all__ = ['core', 'PSFParser', 'PDBParser', 'PQRParser', 'GROParser',
           'CRDParser', 'TOPParser', 'PDBQTParser', 'TPRParser',
           'LAMMPSParser', 'XYZParser', 'GMSParser', 'DLPolyParser',
           'HoomdXMLParser','GSDParser', 'ITPParser']

from . import core
from . import PSFParser
from . import TOPParser
from . import PDBParser
from . import ExtendedPDBParser
from . import PQRParser
from . import GROParser
from . import CRDParser
from . import PDBQTParser
from . import DMSParser
from . import TPRParser
from . import MOL2Parser
from . import LAMMPSParser
from . import XYZParser
from . import TXYZParser
from . import GMSParser
from . import DLPolyParser
from . import HoomdXMLParser
from . import MMTFParser
from . import GSDParser
from . import MinimalParser
from . import ITPParser
from . import FHIAIMSParser



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""
Trajectory transformations --- :mod:`MDAnalysis.transformations`
================================================================

The transformations submodule contains a collection of function-like classes to
modify the trajectory.
Coordinate transformations, such as PBC corrections and molecule fitting
are often required for some analyses and visualization, and the functions in
this module allow transformations to be applied on-the-fly.

A typical transformation class looks like this (note that we keep its name
lowercase because we will treat it as a function, thanks to the ``__call__``
method):

.. code-blocks:: python

    class transformation(object):
        def __init__(self, *args, **kwargs):
            #  do some things
            #  save needed args as attributes.
            self.needed_var = args[0]

        def __call__(self, ts):
            #  apply changes to the Timestep,
            #  or modify an AtomGroup and return Timestep

            return ts

As a concrete example we will write a transformation that rotates a group of
atoms around the z-axis through the center of geometry by a fixed increment
for every time step. We will use
:meth:`MDAnalysis.core.groups.AtomGroup.rotateby`
and simply increment the rotation angle every time the
transformation is called ::

    class spin_atoms(object):
        def __init__(self, atoms, dphi):
            #  Rotate atoms by dphi degrees for every ts around the z axis
            self.atoms = atoms
            self.dphi = dphi
            self.axis = np.array([0, 0, 1])

        def __call__(self, ts):
            phi = self.dphi * ts.frame
            self.atoms.rotateby(phi, self.axis)
            return ts

This transformation can be used as ::

   u = mda.Universe(PSF, DCD)
   u.trajectory.add_transformations(spin_atoms(u.select_atoms("protein"), 1.0))

Also see :mod:`MDAnalysis.transformations.translate` for a simple example.

These transformation functions can be called by the user for any given timestep
of the trajectory, added as a workflow using :meth:`add_transformations`
of the :mod:`~MDAnalysis.coordinates.base`, or upon Universe creation using
the keyword argument `transformations`. Note that in the two latter cases, the
workflow cannot be changed after being defined. for example:

.. code-block:: python

    u = mda.Universe(GRO, XTC)
    trans = transformation(args)
    u.trajectory.add_transformations(trans)

    #  it is equivalent to applying this transforamtion to each Timestep by
    ts = u.trajectory[0]
    ts_trans = trans(ts)

Transformations can also be created as a closure/nested function.
In addition to the specific arguments that each transformation can take, they
also contain a wrapped function that takes a `Timestep` object as argument.
So, a closure-style transformation can be roughly defined as follows:

.. code-block:: python

    def transformation(*args,**kwargs):
        # do some things

            def wrapped(ts):
                #  apply changes to the Timestep,
                #  or modify an AtomGroup and return Timestep

                return ts

            return wrapped

.. Note::
   Although functions (closures) work as transformations, they are not used in
   in MDAnalysis from release 2.0.0 onwards because they cannot be reliably
   serialized and thus a :class:`Universe` with such transformations cannot be
   used with common parallelization schemes (e.g., ones based on
   :mod:`multiprocessing`).
   For detailed descriptions about how to write a closure-style transformation,
   please refer to MDAnalysis 1.x documentation.


.. versionchanged:: 2.0.0
    Transformations should now be created as classes with a :meth:`__call__`
    method instead of being written as a function/closure.
"""

from .base import TransformationBase
from .translate import translate, center_in_box
from .rotate import rotateby
from .positionaveraging import PositionAverager
from .nojump import NoJump
from .fit import fit_translation, fit_rot_trans
from .wrap import wrap, unwrap
from .boxdimensions import set_dimensions



"""Test for MDAnalysis trajectory reader expectations
"""

import importlib
import pickle
import sys
from types import ModuleType
from weakref import ref

import pytest
import numpy as np
from numpy.testing import (
    assert_allclose,
    assert_almost_equal,
    assert_array_almost_equal,
    assert_equal,
)

import MDAnalysis as mda
from MDAnalysis.coordinates.IMD import (
    HAS_IMDCLIENT,
    MIN_IMDCLIENT_VERSION,
    IMDReader,
)
from MDAnalysis.transformations import translate

if HAS_IMDCLIENT:
    import imdclient
    from imdclient.tests.server import InThreadIMDServer
    from imdclient.tests.utils import (
        create_default_imdsinfo_v3,
        get_free_port,
    )

from MDAnalysisTests.coordinates.base import (
    assert_timestep_almost_equal,
    BaseReference,
    MultiframeReaderTest,
)
from MDAnalysisTests.datafiles import (
    COORDINATES_H5MD,
    COORDINATES_TOPOLOGY,
    COORDINATES_TRR,
)


class IMDModuleStateManager:
    """Context manager to completely backup and restore imdclient/IMD module state.

    We need a custom manager because IMD changes its own state (HAS_IMDCLIENT) when it is imported
    and we are going to manipulate the state of the imdclient module that IMD sees.
    """

    def __init__(self):
        self.original_modules = None
        self.imd_was_imported = False

    def __enter__(self):
        # Backup sys.modules
        self.original_modules = sys.modules.copy()

        # Check if IMD module was already imported
        self.imd_was_imported = "MDAnalysis.coordinates.IMD" in sys.modules

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Restore sys.modules completely first
        sys.modules.clear()
        sys.modules.update(self.original_modules)

        # If IMD module was originally imported, force a fresh reload to restore original state
        # This ensures that HAS_IMDCLIENT and other globals are recalculated with the real imdclient
        if self.imd_was_imported:
            # Remove the potentially corrupted IMD module
            sys.modules.pop("MDAnalysis.coordinates.IMD", None)
            # Fresh import will re-evaluate all globals
            import MDAnalysis.coordinates.IMD


class TestImport:
    """Test imdclient import behavior and HAS_IMDCLIENT flag."""

    def _setup_mock_imdclient(self, monkeypatch, version):
        """Helper method to set up mock imdclient with specified version."""
        # Remove IMD and imdclient modules to force fresh import
        monkeypatch.delitem(
            sys.modules, "MDAnalysis.coordinates.IMD", raising=False
        )
        monkeypatch.delitem(sys.modules, "imdclient", raising=False)

        module_name = "imdclient"
        mocked_module = ModuleType(module_name)
        IMDClient_module = ModuleType(f"{module_name}.IMDClient")

        class MockIMDClient:
            pass

        IMDClient_module.IMDClient = MockIMDClient
        mocked_module.IMDClient = IMDClient_module
        mocked_module.__version__ = version

        utils_module = ModuleType(f"{module_name}.utils")
        utils_module.parse_host_port = lambda x: ("localhost", 12345)
        mocked_module.utils = utils_module

        monkeypatch.setitem(sys.modules, module_name, mocked_module)
        monkeypatch.setitem(
            sys.modules, f"{module_name}.IMDClient", IMDClient_module
        )
        monkeypatch.setitem(sys.modules, f"{module_name}.utils", utils_module)

        return mocked_module

    def test_has_minversion(self, monkeypatch):
        """Test that HAS_IMDCLIENT is True when imdclient >= MIN_IMDCLIENT_VERSION."""
        with IMDModuleStateManager():
            self._setup_mock_imdclient(monkeypatch, str(MIN_IMDCLIENT_VERSION))

            # Import and check HAS_IMDCLIENT with compatible version
            import MDAnalysis.coordinates.IMD
            from MDAnalysis.coordinates.IMD import HAS_IMDCLIENT

            assert (
                HAS_IMDCLIENT
            ), f"HAS_IMDCLIENT should be True with version {MIN_IMDCLIENT_VERSION}"

    def test_no_minversion(self, monkeypatch):
        """Test that HAS_IMDCLIENT is False when imdclient version is too old."""
        with IMDModuleStateManager():
            self._setup_mock_imdclient(monkeypatch, "0.0.0")

            # Import and check HAS_IMDCLIENT with incompatible version
            import MDAnalysis.coordinates.IMD
            from MDAnalysis.coordinates.IMD import HAS_IMDCLIENT

            assert (
                not HAS_IMDCLIENT
            ), "HAS_IMDCLIENT should be False with version 0.0.0"

    def test_missing_ImportError(self, monkeypatch):
        """Test that IMDReader raises ImportError when HAS_IMDCLIENT=False."""
        with IMDModuleStateManager():
            self._setup_mock_imdclient(monkeypatch, "0.0.0")

            # Import with incompatible version (HAS_IMDCLIENT=False)
            import MDAnalysis.coordinates.IMD
            from MDAnalysis.coordinates.IMD import IMDReader

            # IMDReader should raise ImportError when HAS_IMDCLIENT=False
            with pytest.raises(
                ImportError, match="IMDReader requires the imdclient"
            ):
                IMDReader("imd://localhost:12345", n_atoms=5)


@pytest.mark.skipif(not HAS_IMDCLIENT, reason="IMDClient not installed")
class IMDReference(BaseReference):
    def __init__(self):
        super(IMDReference, self).__init__()
        # Serve TRR traj data via the server
        traj = mda.coordinates.TRR.TRRReader(COORDINATES_TRR)
        self.server = InThreadIMDServer(traj)
        self.server.set_imdsessioninfo(create_default_imdsinfo_v3())

        self.n_atoms = traj.n_atoms
        self.prec = 3

        self.trajectory = "imd://localhost"
        self.topology = COORDINATES_TOPOLOGY
        self.changing_dimensions = True
        self.reader = IMDReader

        self.first_frame.velocities = self.first_frame.positions / 10
        self.first_frame.forces = self.first_frame.positions / 100

        self.second_frame.velocities = self.second_frame.positions / 10
        self.second_frame.forces = self.second_frame.positions / 100

        self.last_frame.velocities = self.last_frame.positions / 10
        self.last_frame.forces = self.last_frame.positions / 100

        self.jump_to_frame.velocities = self.jump_to_frame.positions / 10
        self.jump_to_frame.forces = self.jump_to_frame.positions / 100

    def iter_ts(self, i):
        ts = self.first_frame.copy()
        ts.positions = 2**i * self.first_frame.positions
        ts.velocities = ts.positions / 10
        ts.forces = ts.positions / 100
        ts.time = i
        ts.frame = i
        return ts


@pytest.mark.skipif(not HAS_IMDCLIENT, reason="IMDClient not installed")
class TestIMDReaderBaseAPI(MultiframeReaderTest):

    @pytest.fixture(scope="function")
    def ref(self):
        """Not a static method like in base class- need new server for each test"""
        reference = IMDReference()
        yield reference
        reference.server.cleanup()

    @staticmethod
    @pytest.fixture()
    def reader(ref):
        # This will start the test IMD Server, waiting for a connection
        # to then send handshake & first frame
        ref.server.handshake_sequence("localhost")
        # This will connect to the test IMD Server and read the first frame
        reader = ref.reader(
            f"{ref.trajectory}:{ref.server.port}",
            n_atoms=ref.n_atoms,
            buffer_size=1 * 1024 * 1024,
        )
        # Send the rest of the frames- small enough to all fit in socket itself
        ref.server.send_frames(1, 5)

        reader.add_auxiliary(
            "lowf",
            ref.aux_lowf,
            dt=ref.aux_lowf_dt,
            initial_time=0,
            time_selector=None,
        )
        reader.add_auxiliary(
            "highf",
            ref.aux_highf,
            dt=ref.aux_highf_dt,
            initial_time=0,
            time_selector=None,
        )
        yield reader
        reader.close()

    @staticmethod
    @pytest.fixture()
    def transformed(ref):
        # This will start the test IMD Server, waiting for a connection
        # to then send handshake & first frame
        ref.server.handshake_sequence("localhost")
        # This will connect to the test IMD Server and read the first frame
        transformed = ref.reader(
            f"{ref.trajectory}:{ref.server.port}",
            n_atoms=ref.n_atoms,
            buffer_size=1 * 1024 * 1024,
        )
        # Send the rest of the frames- small enough to all fit in socket itself
        ref.server.send_frames(1, 5)
        transformed.add_transformations(
            translate([1, 1, 1]), translate([0, 0, 0.33])
        )
        return transformed

    def test_n_frames(self, ref, reader):
        with pytest.raises(RuntimeError, match="n_frames is unknown"):
            reader.n_frames

    def test_first_frame(self, ref, reader):
        # don't rewind here as in inherited base test
        assert_timestep_almost_equal(
            reader.ts, ref.first_frame, decimal=ref.prec
        )

    def test_get_writer_1(self, ref, reader, tmpdir):
        with pytest.raises(
            RuntimeError,
            match="cannot create Writer for streamed trajectories",
        ):
            reader.Writer(str(tmpdir.join("output")))

    def test_get_writer_2(self, ref, reader, tmpdir):
        with pytest.raises(
            RuntimeError,
            match="cannot create Writer for streamed trajectories",
        ):
            reader.Writer(str(tmpdir.join("output")), n_atoms=100)

    def test_OtherWriter_RuntimeError(self, reader, tmpdir):
        with pytest.raises(
            RuntimeError,
            match="cannot create OtherWriter for streamed trajectories",
        ):
            reader.OtherWriter(tmpdir.join("output"))

    def test_total_time(self, ref, reader):
        pytest.skip("`total_time` is unknown for IMDReader")

    def test_changing_dimensions(self, ref, reader):
        with pytest.raises(
            RuntimeError, match="Stream-based readers can't be rewound"
        ):
            reader.rewind()

    def test_iter(self, ref, reader):
        for i, ts in enumerate(reader):
            assert_timestep_almost_equal(ts, ref.iter_ts(i), decimal=ref.prec)

    def test_first_dimensions(self, ref, reader):
        # don't rewind here as in inherited base test
        if ref.dimensions is None:
            assert reader.ts.dimensions is None
        else:
            assert_allclose(
                reader.ts.dimensions,
                ref.dimensions,
                rtol=0,
                atol=1.5 * 10 ** (-ref.prec),
            )

    def test_transformed(self, ref, transformed):
        # see transformed fixture
        ref_trans = ref.first_frame.positions + 1
        ref_trans[:, 2] += 0.33
        assert_allclose(transformed.ts.positions, ref_trans)

    def test_volume(self, ref, reader):
        # don't rewind here as in inherited base test
        vol = reader.ts.volume
        # Here we can only be sure about the numbers upto the decimal point due
        # to limited floating point precision.
        assert_allclose(vol, ref.volume, rtol=0, atol=1.5e0)

    def test_reload_auxiliaries_from_description(self, ref, reader):
        pytest.skip("Cannot create two IMDReaders on the same stream")

    def test_stop_iter(self, reader):
        with pytest.raises(
            RuntimeError, match="Stream-based readers can't be rewound"
        ):
            reader.rewind()

    def test_iter_rewinds(self, reader):
        pytest.skip("IMDReader cannot be rewound")

    def test_timeseries_shape(self, reader):
        pytest.skip("IMDReader does not support timeseries")

    def test_timeseries_asel_shape(self, reader):
        pytest.skip("IMDReader does not support timeseries")

    def test_timeseries_values(self, reader):
        pytest.skip("IMDReader does not support timeseries")

    def test_transformations_2iter(self, ref, transformed):
        pytest.skip("IMDReader cannot be reopened")

    def test_transformations_slice(self, ref, transformed):
        pytest.skip("IMDReader cannot be reopened")

    def test_transformations_switch_frame(self, ref, transformed):
        pytest.skip("IMDReader cannot be reopened")

    def test_transformation_rewind(self, ref, transformed):
        pytest.skip("IMDReader cannot be reopened")

    def test_pickle_reader(self, reader):
        with pytest.raises(
            NotImplementedError, match="does not support pickling"
        ):
            pickle.dumps(reader)

    def test_pickle_next_ts_reader(self, reader):
        pytest.skip("IMDReader cannot be pickled")

    def test_pickle_last_ts_reader(self, reader):
        pytest.skip("IMDReader cannot be pickled")

    def test_transformations_copy(self, ref, transformed):
        with pytest.raises(RuntimeError, match="does not support copying"):
            transformed.copy()

    def test_timeseries_empty_asel(self, reader):
        pytest.skip("IMDReader does not support timeseries")

    def test_timeseries_empty_atomgroup(self, reader):
        pytest.skip("IMDReader does not support timeseries")

    def test_timeseries_asel_warns_deprecation(self, reader):
        pytest.skip("IMDReader does not support timeseries")

    def test_timeseries_atomgroup(self, reader):
        pytest.skip("IMDReader does not support timeseries")

    def test_timeseries_atomgroup_asel_mutex(self, reader):
        pytest.skip("IMDReader does not support timeseries")

    def test_last_frame(self, ref, reader):
        pytest.skip("IMDReader cannot be rewound")

    def test_go_over_last_frame(self, ref, reader):
        pytest.skip("IMDReader must be an indexed using a slice")

    def test_frame_jump(self, ref, reader):
        pytest.skip("IMDReader must be an indexed using a slice")

    def test_frame_jump_issue1942(self, ref, reader):
        pytest.skip("IMDReader must be an indexed using a slice")

    def test_next_gives_second_frame(self, ref, reader):
        # don't recreate reader here as in inherited base test
        ts = reader.next()
        assert_timestep_almost_equal(ts, ref.second_frame, decimal=ref.prec)

    def test_frame_collect_all_same(self, reader):
        pytest.skip("IMDReader has independent coordinates")


@pytest.fixture
def universe():
    return mda.Universe(COORDINATES_TOPOLOGY, COORDINATES_H5MD)


@pytest.fixture
def port():
    return get_free_port()


@pytest.fixture
def imdsinfo():
    return create_default_imdsinfo_v3()


@pytest.mark.skipif(not HAS_IMDCLIENT, reason="IMDClient not installed")
class TestStreamIteration:
    @pytest.fixture
    def reader(self, universe, imdsinfo):
        server = InThreadIMDServer(universe.trajectory)
        server.set_imdsessioninfo(imdsinfo)
        server.handshake_sequence("localhost", first_frame=True)
        reader = IMDReader(
            f"imd://localhost:{server.port}",
            n_atoms=universe.trajectory.n_atoms,
            buffer_size=1 * 1024 * 1024,
        )
        server.send_frames(1, 5)

        yield reader
        server.cleanup()
        reader.close()

    def test_iterate_step(self, reader, universe):
        i = 0
        for ts in reader[::2]:
            assert ts.frame == i
            i += 2

    def test_iterate_twice_sliced_raises_error(self, reader):
        for ts in reader[::2]:
            pass
        with pytest.raises(RuntimeError, match="Cannot reopen stream"):
            for ts in reader[::2]:
                pass

    def test_iterate_twice_all_raises_error(self, reader):
        for ts in reader:
            pass
        with pytest.raises(RuntimeError, match="Cannot reopen stream"):
            for ts in reader:
                pass

    def test_iterate_twice_fi_all_raises_error(self, reader):
        for ts in reader[:]:
            pass
        with pytest.raises(RuntimeError, match="Cannot reopen stream"):
            for ts in reader[:]:
                pass

    def test_index_stream_raises_error(self, reader):
        with pytest.raises(TypeError, match="Streamed trajectories must be"):
            reader[0]

    def test_iterate_backwards_raises_error(self, reader):
        with pytest.raises(ValueError, match="Cannot go backwards"):
            for ts in reader[::-1]:
                pass

    def test_iterate_start_stop_raises_error(self, reader):
        with pytest.raises(ValueError, match="Cannot expect a start index"):
            for ts in reader[1:3]:
                pass

    def test_subslice_fi_all_after_iteration_raises_error(self, reader):
        sliced_reader = reader[:]
        for ts in sliced_reader:
            pass
        sub_sliced_reader = sliced_reader[::1]
        with pytest.raises(RuntimeError):
            for ts in sub_sliced_reader:
                pass

    def test_timeseries_raises(self, reader):
        with pytest.raises(
            RuntimeError,
            match="cannot access timeseries for streamed trajectories",
        ):
            reader.timeseries()

    def test_step_property(self, reader):
        """Test that the step property returns the correct step size."""
        # Test step property for different slice steps
        sliced_reader = reader[::1]
        assert sliced_reader.step == 1

        sliced_reader_step5 = reader[::5]
        assert sliced_reader_step5.step == 5


@pytest.mark.skipif(not HAS_IMDCLIENT, reason="IMDClient not installed")
def test_n_atoms_not_specified(universe, imdsinfo):
    server = InThreadIMDServer(universe.trajectory)
    server.set_imdsessioninfo(imdsinfo)
    server.handshake_sequence("localhost", first_frame=True)
    with pytest.raises(
        ValueError,
        match="IMDReader: n_atoms must be specified",
    ):
        IMDReader(
            f"imd://localhost:{server.port}",
        )
    server.cleanup()


@pytest.mark.skipif(not HAS_IMDCLIENT, reason="IMDClient not installed")
def test_imd_stream_empty(universe, imdsinfo):
    server = InThreadIMDServer(universe.trajectory)
    server.set_imdsessioninfo(imdsinfo)
    server.handshake_sequence("localhost", first_frame=False)
    with pytest.raises(
        RuntimeError,
        match="IMDReader: Read error",
    ):
        IMDReader(
            f"imd://localhost:{server.port}",
            n_atoms=universe.trajectory.n_atoms,
        )
    server.cleanup()


@pytest.mark.skipif(not HAS_IMDCLIENT, reason="IMDClient not installed")
def test_create_imd_universe(universe, imdsinfo):
    server = InThreadIMDServer(universe.trajectory)
    server.set_imdsessioninfo(imdsinfo)
    server.handshake_sequence("localhost", first_frame=True)
    u_imd = mda.Universe(
        COORDINATES_TOPOLOGY,
        f"imd://localhost:{server.port}",
        n_atoms=universe.trajectory.n_atoms,
    )
    assert type(u_imd.trajectory).__name__ == "IMDReader"
    with pytest.raises(ValueError, match="IMDReader: Invalid IMD URL"):
        u_imd = mda.Universe(
            COORDINATES_TOPOLOGY,
            f"imd://localhost:{port}/invalid",
            n_atoms=universe.trajectory.n_atoms,
        )
    server.cleanup()


def test_imd_format_hint():
    assert IMDReader._format_hint("imd://localhost:12345")
    assert IMDReader._format_hint("imd://localhost:12345/invalid")
    assert not IMDReader._format_hint("not_a_valid_imd_url")
    assert not IMDReader._format_hint(12345)
    assert not IMDReader._format_hint(None)


@pytest.mark.skipif(not HAS_IMDCLIENT, reason="IMDClient not installed")
def test_wrong_imd_protocol_version(universe, imdsinfo):
    """Test that IMDReader raises ValueError for non-v3 protocol versions."""
    # Modify the fixture to have wrong version
    imdsinfo.version = 2  # Wrong version, should be 3

    server = InThreadIMDServer(universe.trajectory)
    server.set_imdsessioninfo(imdsinfo)
    server.handshake_sequence("localhost", first_frame=True)

    with pytest.raises(
        ValueError,
        match=rf"IMDReader: Detected IMD version v{imdsinfo.version}, "
        rf"but IMDReader is only compatible with v3",
    ):
        IMDReader(
            f"imd://localhost:{server.port}",
            n_atoms=universe.trajectory.n_atoms,
        )
    server.cleanup()



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest

import MDAnalysis as mda
import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_equal,
)

from MDAnalysis.coordinates.XYZ import XYZWriter

from MDAnalysisTests.datafiles import COORDINATES_XYZ, COORDINATES_XYZ_BZ2
from MDAnalysisTests.coordinates.base import (
    MultiframeReaderTest,
    BaseReference,
    BaseWriterTest,
)
from MDAnalysisTests import make_Universe

from MDAnalysis import __version__


class XYZReference(BaseReference):
    def __init__(self):
        super(XYZReference, self).__init__()
        self.trajectory = COORDINATES_XYZ
        # XYZ is it's own Topology
        self.topology = COORDINATES_XYZ
        self.reader = mda.coordinates.XYZ.XYZReader
        self.writer = mda.coordinates.XYZ.XYZWriter
        self.ext = "xyz"
        self.volume = 0
        self.dimensions = None
        self.container_format = True


class TestXYZReader(MultiframeReaderTest):
    @staticmethod
    @pytest.fixture()
    def ref():
        return XYZReference()

    def test_double_open(self):
        with pytest.raises(Exception):
            self.reader.open_trajectory()
            self.reader.open_trajectory()


class TestXYZWriter(BaseWriterTest):
    @staticmethod
    @pytest.fixture()
    def ref():
        return XYZReference()

    def test_write_selection(self, ref, universe, reader, tmpdir):
        sel_str = "name CA"
        sel = universe.select_atoms(sel_str)
        outfile = "write-selection-test" + ref.ext

        with tmpdir.as_cwd():
            with ref.writer(outfile, sel.n_atoms) as W:
                for ts in universe.trajectory:
                    W.write(sel.atoms)

            copy = ref.reader(outfile)
            for orig_ts, copy_ts in zip(universe.trajectory, copy):
                assert_almost_equal(
                    copy_ts._pos,
                    sel.atoms.positions,
                    ref.prec,
                    err_msg="coordinate mismatch between original and written "
                    "trajectory at frame {} (orig) vs {} (copy)".format(
                        orig_ts.frame, copy_ts.frame
                    ),
                )

    def test_write_different_models_in_trajectory(self, ref, universe, tmpdir):
        outfile = "write-models-in-trajectory" + ref.ext
        # n_atoms should match for each TimeStep if it was specified
        with tmpdir.as_cwd():
            with ref.writer(outfile, n_atoms=4) as w:
                with pytest.raises(ValueError):
                    w.write(universe)

    def test_no_conversion(self, ref, universe, reader, tmpdir):
        outfile = "write-no-conversion" + ref.ext
        with tmpdir.as_cwd():
            with ref.writer(outfile, convert_units=False) as w:
                for ts in universe.trajectory:
                    w.write(universe)
            self._check_copy(outfile, ref, reader)

    @pytest.mark.parametrize(
        "remarkout, remarkin",
        [
            2 * ["Curstom Remark"],
            2 * [""],
            [
                None,
                "frame 0 | Written by MDAnalysis XYZWriter (release {0})".format(
                    __version__
                ),
            ],
        ],
    )
    def test_remark(self, universe, remarkout, remarkin, ref, tmpdir):
        outfile = "write-remark.xyz"

        with tmpdir.as_cwd():
            universe.atoms.write(outfile, remark=remarkout)

            with open(outfile, "r") as xyzout:
                lines = xyzout.readlines()

                assert lines[1].strip() == remarkin

    def test_precision(self, universe, tmpdir):
        outfile = "write-precision.xyz"
        precision = 10

        with tmpdir.as_cwd():
            universe.atoms.write(outfile, precision=precision)
            with open(outfile, "r") as xyzout:
                lines = xyzout.readlines()
                # check that the precision is set correctly
                assert len(lines[2].split()[1].split(".")[1]) == precision


class XYZ_BZ_Reference(XYZReference):
    def __init__(self):
        super(XYZ_BZ_Reference, self).__init__()
        self.trajectory = COORDINATES_XYZ_BZ2
        self.ext = "xyz.bz2"


class Test_XYZBZReader(TestXYZReader):
    @staticmethod
    @pytest.fixture()
    def ref():
        return XYZ_BZ_Reference()


class Test_XYZBZWriter(TestXYZWriter):
    @staticmethod
    @pytest.fixture()
    def ref():
        return XYZ_BZ_Reference()


class TestXYZWriterNames(object):
    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir.join("/outfile.xyz"))

    def test_no_names(self, outfile):
        """Atoms should all be named 'X'"""
        u = make_Universe(trajectory=True)

        wmsg = "does not have atom elements or names"

        with pytest.warns(UserWarning, match=wmsg):
            w = XYZWriter(outfile)
            w.write(u)
            w.close()

        u2 = mda.Universe(outfile)
        assert all(u2.atoms.names == "X")

    @pytest.mark.parametrize("attr", ["elements", "names"])
    def test_elements_or_names(self, outfile, attr):

        u = mda.Universe.empty(n_atoms=5, trajectory=True)

        u.add_TopologyAttr(attr, values=["Te", "S", "Ti", "N", "Ga"])

        with mda.Writer(outfile) as w:
            w.write(u)

        with open(outfile, "r") as r:
            names = "".join(l.split()[0].strip() for l in r.readlines()[2:-1])

        assert names[:-1].lower() == "testing"

    def test_elements_and_names(self, outfile):
        """Should always default to elements over names"""
        u = mda.Universe.empty(n_atoms=5, trajectory=True)

        u.add_TopologyAttr("elements", values=["Te", "S", "Ti", "N", "Ga"])
        u.add_TopologyAttr("names", values=["A", "B", "C", "D", "E"])

        with mda.Writer(outfile) as w:
            w.write(u)

        u2 = mda.Universe(outfile)
        assert_equal(u2.atoms.names, u.atoms.elements)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest

from MDAnalysisTests.datafiles import (
    TRZ,
    TRZ_psf,
    waterPSF,
    waterDCD,
    XYZ_mini,
)
from numpy.testing import assert_almost_equal
import numpy as np
from unittest import mock
from importlib import reload

import MDAnalysis as mda
from MDAnalysis.analysis import hbonds
from MDAnalysis.analysis.hbonds import HydrogenBondAutoCorrel as HBAC


@pytest.fixture(scope="module")
def u_water():
    return mda.Universe(waterPSF, waterDCD)


class TestHydrogenBondAutocorrel(object):
    @pytest.fixture()
    def u(self):
        return mda.Universe(TRZ_psf, TRZ)

    @pytest.fixture()
    def hydrogens(self, u):
        return u.atoms.select_atoms("name Hn")

    @pytest.fixture()
    def nitrogens(self, u):
        return u.atoms.select_atoms("name N")

    @pytest.fixture()
    def oxygens(self, u):
        return u.atoms.select_atoms("name O")

    # regression tests for different conditions
    def test_continuous(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(
            u,
            hydrogens=hydrogens,
            acceptors=oxygens,
            donors=nitrogens,
            bond_type="continuous",
            sample_time=0.06,
        )
        hbond.run()

        assert_almost_equal(
            hbond.solution["results"],
            np.array(
                [
                    1.0,
                    0.92668623,
                    0.83137828,
                    0.74486804,
                    0.67741936,
                    0.60263932,
                ],
                dtype=np.float32,
            ),
        )

    def test_continuous_excl(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(
            u,
            hydrogens=hydrogens,
            acceptors=oxygens,
            donors=nitrogens,
            bond_type="continuous",
            exclusions=(
                np.arange(len(hydrogens)),
                np.array(range(len(oxygens))),
            ),
            sample_time=0.06,
        )
        hbond.run()

        assert_almost_equal(
            hbond.solution["results"],
            np.array(
                [
                    1.0,
                    0.92668623,
                    0.83137828,
                    0.74486804,
                    0.67741936,
                    0.60263932,
                ],
                dtype=np.float32,
            ),
        )

    def test_intermittent(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(
            u,
            hydrogens=hydrogens,
            acceptors=oxygens,
            donors=nitrogens,
            bond_type="intermittent",
            sample_time=0.06,
        )
        hbond.run()

        assert_almost_equal(
            hbond.solution["results"],
            np.array(
                [
                    1.0,
                    0.92668623,
                    0.84310848,
                    0.79325515,
                    0.76392961,
                    0.72287393,
                ],
                dtype=np.float32,
            ),
        )

    def test_intermittent_timecut(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(
            u,
            hydrogens=hydrogens,
            acceptors=oxygens,
            donors=nitrogens,
            bond_type="intermittent",
            time_cut=0.01,  # time cut at traj.dt == continuous
            sample_time=0.06,
        )
        hbond.run()

        assert_almost_equal(
            hbond.solution["results"],
            np.array(
                [
                    1.0,
                    0.92668623,
                    0.83137828,
                    0.74486804,
                    0.67741936,
                    0.60263932,
                ],
                dtype=np.float32,
            ),
        )

    def test_intermittent_excl(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(
            u,
            hydrogens=hydrogens,
            acceptors=oxygens,
            donors=nitrogens,
            bond_type="intermittent",
            exclusions=(
                np.arange(len(hydrogens)),
                np.array(range(len(oxygens))),
            ),
            sample_time=0.06,
        )
        hbond.run()

        assert_almost_equal(
            hbond.solution["results"],
            np.array(
                [
                    1.0,
                    0.92668623,
                    0.84310848,
                    0.79325515,
                    0.76392961,
                    0.72287393,
                ],
                dtype=np.float32,
            ),
        )

    # For `solve` the test trajectories aren't long enough
    # So spoof the results and check that solver finds solution
    def test_solve_continuous(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(
            u,
            hydrogens=hydrogens,
            acceptors=oxygens,
            donors=nitrogens,
            bond_type="continuous",
            sample_time=0.06,
        )

        def actual_function_cont(t):
            A1 = 0.75
            A2 = 0.25
            tau1 = 0.5
            tau2 = 0.1
            return A1 * np.exp(-t / tau1) + A2 * np.exp(-t / tau2)

        hbond.solution["time"] = time = np.arange(0, 0.06, 0.001)
        hbond.solution["results"] = actual_function_cont(time)

        hbond.solve()

        assert_almost_equal(
            hbond.solution["fit"],
            np.array([0.75, 0.5, 0.1]),
        )

    def test_solve_intermittent(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(
            u,
            hydrogens=hydrogens,
            acceptors=oxygens,
            donors=nitrogens,
            bond_type="intermittent",
            sample_time=0.06,
        )

        def actual_function_int(t):
            A1 = 0.33
            A2 = 0.33
            A3 = 0.34
            tau1 = 5
            tau2 = 1
            tau3 = 0.1
            return (
                A1 * np.exp(-t / tau1)
                + A2 * np.exp(-t / tau2)
                + A3 * np.exp(-t / tau3)
            )

        hbond.solution["time"] = time = np.arange(0, 6.0, 0.01)
        hbond.solution["results"] = actual_function_int(time)

        hbond.solve()

        assert_almost_equal(
            hbond.solution["fit"],
            np.array([0.33, 0.33, 5, 1, 0.1]),
        )

    # setup errors
    def test_wronglength_DA(self, u, hydrogens, oxygens, nitrogens):
        with pytest.raises(ValueError):
            HBAC(
                u,
                hydrogens=hydrogens[:-1],
                acceptors=oxygens,
                donors=nitrogens,
                bond_type="intermittent",
                exclusions=(
                    np.arange(len(hydrogens)),
                    np.array(range(len(oxygens))),
                ),
                sample_time=0.06,
            )

    def test_exclusions(self, u, hydrogens, oxygens, nitrogens):
        excl_list = (
            np.array(range(len(hydrogens))),
            np.array(range(len(oxygens))),
        )
        excl_list2 = excl_list[0], excl_list[1][:-1]
        with pytest.raises(ValueError):
            HBAC(
                u,
                hydrogens=hydrogens,
                acceptors=oxygens,
                donors=nitrogens,
                bond_type="intermittent",
                exclusions=excl_list2,
                sample_time=0.06,
            )

    def test_bond_type_VE(self, u, hydrogens, oxygens, nitrogens):
        with pytest.raises(ValueError):
            HBAC(
                u,
                hydrogens=hydrogens,
                acceptors=oxygens,
                donors=nitrogens,
                bond_type="marzipan",
                exclusions=(
                    np.arange(len(hydrogens)),
                    np.array(range(len(oxygens))),
                ),
                sample_time=0.06,
            )

    def test_solve_before_run_VE(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(
            u,
            hydrogens=hydrogens,
            acceptors=oxygens,
            donors=nitrogens,
            bond_type="continuous",
            sample_time=0.06,
        )
        with pytest.raises(ValueError):
            hbond.solve()

    @mock.patch("MDAnalysis.coordinates.TRZ.TRZReader._read_frame")
    def test_unslicable_traj_VE(
        self, mock_read, u, hydrogens, oxygens, nitrogens
    ):
        mock_read.side_effect = TypeError

        with pytest.raises(ValueError):
            HBAC(
                u,
                hydrogens=hydrogens,
                acceptors=oxygens,
                donors=nitrogens,
                bond_type="continuous",
                sample_time=0.06,
            )

    def test_repr(self, u, hydrogens, oxygens, nitrogens):
        hbond = HBAC(
            u,
            hydrogens=hydrogens,
            acceptors=oxygens,
            donors=nitrogens,
            bond_type="continuous",
            sample_time=0.06,
        )
        assert isinstance(repr(hbond), str)

    def test_deprecation_warning(self, u, hydrogens, oxygens, nitrogens):
        wmsg = (
            "`HydrogenBondAutoCorrel` is deprecated!\n"
            "`HydrogenBondAutoCorrel` will be removed in release 3.0.0.\n"
            "The class was moved to MDAnalysis.analysis.hbonds.hbond_autocorrel."
        )
        with pytest.warns(DeprecationWarning, match=wmsg):
            HBAC(
                u,
                hydrogens=hydrogens,
                acceptors=oxygens,
                donors=nitrogens,
                bond_type="continuous",
                sample_time=0.06,
            )


def test_find_donors(u_water):
    H = u_water.select_atoms("name H*")

    D = hbonds.find_hydrogen_donors(H)

    assert len(H) == len(D)
    # check each O is bonded to the corresponding H
    for h_atom, o_atom in zip(H, D):
        assert o_atom in h_atom.bonded_atoms


def test_donors_nobonds():
    u = mda.Universe(XYZ_mini)

    with pytest.raises(mda.NoDataError):
        hbonds.find_hydrogen_donors(u.atoms)


def test_find_hydrogen_donors_deprecation_warning(u_water):
    H = u_water.select_atoms("name H*")
    wmsg = (
        "`find_hydrogen_donors` is deprecated!\n"
        "`find_hydrogen_donors` will be removed in release 3.0.0.\n"
        "The function was moved to MDAnalysis.analysis.hbonds.hbond_autocorrel."
    )
    with pytest.warns(DeprecationWarning, match=wmsg):
        hbonds.find_hydrogen_donors(H)


def test_moved_module_warning():
    wmsg = (
        "This module was moved to "
        "MDAnalysis.analysis.hydrogenbonds.hbond_autocorrel; "
        "hbonds.hbond_autocorrel will be removed in 3.0.0."
    )
    with pytest.warns(DeprecationWarning, match=wmsg):
        reload(hbonds.hbond_autocorrel)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_equal,
)
import pytest

import MDAnalysis as mda
from MDAnalysis.lib.distances import calc_bonds, calc_angles, calc_dihedrals
from MDAnalysisTests.datafiles import LAMMPSdata_many_bonds
from MDAnalysis.core.topologyobjects import (
    TopologyGroup,
    TopologyObject,
    TopologyDict,
    # TODO: the following items are not used
    Bond,
    Angle,
    Dihedral,
    ImproperDihedral,
)


from MDAnalysisTests.datafiles import PSF, DCD, TRZ_psf, TRZ


@pytest.fixture(scope="module")
def PSFDCD():
    return mda.Universe(PSF, DCD)


class TestTopologyObjects(object):
    """Test the base TopologyObject funtionality

    init
    repr
    eq
    ne
    iter
    len
    """

    precision = 3  # see Issue #271 and #1556

    @staticmethod
    @pytest.fixture
    def a1(PSFDCD):
        return PSFDCD.atoms[1:3]

    @staticmethod
    @pytest.fixture
    def a2(PSFDCD):
        return PSFDCD.atoms[3:5]

    @staticmethod
    @pytest.fixture
    def TO1(a1):
        return TopologyObject(a1.indices, a1.universe)

    @staticmethod
    @pytest.fixture
    def TO2(a2):
        return TopologyObject(a2.indices, a2.universe)

    @staticmethod
    @pytest.fixture
    def b(PSFDCD):
        return PSFDCD.atoms[12].bonds[0]

    def test_repr(self, TO1):
        assert_equal(repr(TO1), "<TopologyObject between: Atom 1, Atom 2>")

    def test_eq(self, a1, TO1, TO2, PSFDCD):
        TO1_b = TopologyObject(a1.indices, PSFDCD)

        assert_equal(TO1 == TO1_b, True)
        assert_equal(TO1 == TO2, False)

    def test_ne(self, TO1, TO2, a1, PSFDCD):
        TO1_b = TopologyObject(a1.indices, PSFDCD)

        assert_equal(TO1 != TO1_b, False)
        assert_equal(TO1 != TO2, True)

    def test_gt(self, TO1, TO2):
        assert_equal(TO1 > TO2, False)

    def test_lt(self, TO1, TO2):
        assert_equal(TO1 < TO2, True)

    def test_iter(self, a1, TO1):
        assert_equal(list(a1), list(TO1))

    def test_len(self, a1):
        assert_equal(len(a1), 2)

    def test_hash(self, TO1, TO2, a1, PSFDCD):
        assert hash(TO1) != hash(TO2)

        # Different universe should yield different hash
        u = mda.Universe(PSF, DCD)
        ag = u.atoms[1:3]
        TO3 = TopologyObject(ag.indices, u)
        assert hash(TO1) != hash(TO3)

        # Different order should yield different hash
        u = mda.Universe(PSF, DCD)
        ag = u.atoms[[2, 1]]
        TO3 = TopologyObject(ag.indices, u)
        assert hash(TO1) != hash(TO3)

        # should work with TO from the same atomgroup
        TO3 = TopologyObject(a1.indices, PSFDCD)
        assert_equal(hash(TO1), hash(TO3))

    def test_indices(self, b):
        assert_equal(b.indices, tuple([a.index for a in b.atoms]))

    # Bond class checks
    def test_partner(self, b):
        a1, a2 = b
        assert_equal(b.partner(a1), a2)
        assert_equal(b.partner(a2), a1)

    def test_partner_VE(self, PSFDCD, b):
        a3 = PSFDCD.atoms[0]
        with pytest.raises(ValueError):
            b.partner(a3)

    def test_bondlength(self, b):
        assert_almost_equal(b.length(), 1.7661301556941993, self.precision)

    def test_bondrepr(self, b):
        assert_equal(repr(b), "<Bond between: Atom 9, Atom 12>")

    # Angle class checks
    def test_angle(self, PSFDCD):
        angle = PSFDCD.atoms[210].angles[0]

        assert_almost_equal(angle.angle(), 107.20893, self.precision)
        assert_almost_equal(angle.value(), 107.20893, self.precision)

    def test_angle_repr(self, PSFDCD):
        angle = PSFDCD.atoms[[30, 10, 20]].angle

        assert_equal(repr(angle), "<Angle between: Atom 30, Atom 10, Atom 20>")

    def test_angle_180(self):
        # we edit the coordinates, so make our own universe
        u = mda.Universe(PSF, DCD)
        angle = u.atoms[210].angles[0]
        coords = np.array([[1, 1, 1], [2, 1, 1], [3, 1, 1]], dtype=np.float32)

        angle.atoms.positions = coords

        assert_almost_equal(angle.value(), 180.0, self.precision)

    # Dihedral class check
    def test_dihedral(self, PSFDCD):
        dihedral = PSFDCD.atoms[14].dihedrals[0]

        assert_almost_equal(dihedral.dihedral(), 18.317778, self.precision)
        assert_almost_equal(dihedral.value(), 18.317778, self.precision)

    def test_dihedral_repr(self, PSFDCD):
        dihedral = PSFDCD.atoms[[4, 7, 8, 1]].dihedral

        assert_equal(
            repr(dihedral),
            "<Dihedral between: Atom 4, Atom 7, Atom 8, Atom 1>",
        )

    # Improper_Dihedral class check
    def test_improper(self, PSFDCD):
        imp = PSFDCD.atoms[4].impropers[0]

        assert_almost_equal(imp.improper(), -3.8370631, self.precision)
        assert_almost_equal(imp.value(), -3.8370631, self.precision)

    def test_improper_repr(self, PSFDCD):
        imp = PSFDCD.atoms[[4, 7, 8, 1]].improper

        assert_equal(
            repr(imp),
            "<ImproperDihedral between: Atom 4, Atom 7, Atom 8, Atom 1>",
        )

    def test_ureybradley_repr(self, PSFDCD):
        ub = PSFDCD.atoms[[30, 10]].ureybradley

        assert_equal(repr(ub), "<UreyBradley between: Atom 30, Atom 10>")

    def test_ureybradley_repr_VE(self, PSFDCD):
        with pytest.raises(ValueError):
            ub = PSFDCD.atoms[[30, 10, 2]].ureybradley

    def test_ureybradley_partner(self, PSFDCD):
        ub = PSFDCD.atoms[[30, 10]].ureybradley
        assert ub.partner(PSFDCD.atoms[30]) == PSFDCD.atoms[10]
        assert ub.partner(PSFDCD.atoms[10]) == PSFDCD.atoms[30]

    def test_ureybradley_distance(self, b):
        assert_almost_equal(
            b.atoms.ureybradley.distance(), b.length(), self.precision
        )

    def test_cmap_repr(self, PSFDCD):
        cmap = PSFDCD.atoms[[4, 7, 8, 1, 2]].cmap

        assert_equal(
            repr(cmap),
            "<CMap between: Atom 4, Atom 7, Atom 8, Atom 1, Atom 2>",
        )

    def test_cmap_repr_VE(self, PSFDCD):
        with pytest.raises(ValueError):
            cmap = PSFDCD.atoms[[30, 10, 2]].cmap


class TestTopologyGroup(object):
    """Tests TopologyDict and TopologyGroup classes with psf input"""

    @staticmethod
    @pytest.fixture
    def res1(PSFDCD):
        return PSFDCD.residues[0]

    @staticmethod
    @pytest.fixture
    def res2(PSFDCD):
        return PSFDCD.residues[1]

    @staticmethod
    @pytest.fixture
    def b_td(PSFDCD):
        return PSFDCD.atoms.bonds.topDict

    @staticmethod
    @pytest.fixture
    def a_td(PSFDCD):
        return PSFDCD.atoms.angles.topDict

    @staticmethod
    @pytest.fixture
    def t_td(PSFDCD):
        return PSFDCD.atoms.dihedrals.topDict

    # Checking TopologyDict functionality
    # * check that enough types are made
    # * check the identity of one of the keys
    # * check uniqueness of keys (ie reversed doesnt exist)
    # * then check same key reversed is accepted
    # * then select based on key
    # * then select based on reversed key and check output is the same
    # all for Bonds Angles and Dihedrals
    def test_td_len(self, b_td):
        assert len(b_td) == 57

    def test_td_iter(self, b_td):
        assert list(b_td) == list(b_td.dict.keys())

    def test_td_keyerror(self, b_td):
        with pytest.raises(KeyError):
            b_td[("something", "stupid")]

    def test_td_universe(self, b_td, PSFDCD):
        assert b_td.universe is PSFDCD

    def test_bonds_types(self, PSFDCD, res1):
        """Tests TopologyDict for bonds"""
        assert len(PSFDCD.atoms.bonds.types()) == 57
        assert len(res1.atoms.bonds.types()) == 12

    def test_bonds_contains(self, b_td):
        assert ("57", "2") in b_td

    def test_bond_uniqueness(self, PSFDCD):
        bondtypes = PSFDCD.atoms.bonds.types()
        # check that a key doesn't appear in reversed format in keylist
        # have to exclude case of b[::-1] == b as this is false positive
        assert not any(
            [b[::-1] in bondtypes for b in bondtypes if b[::-1] != b]
        )

    def test_bond_reversal(self, PSFDCD, b_td):
        bondtypes = PSFDCD.atoms.bonds.types()
        b = bondtypes[1]
        assert all([b in b_td, b[::-1] in b_td])

        tg1 = b_td[b]
        tg2 = b_td[b[::-1]]
        assert tg1 == tg2

    # This test will pass as long as `TopologyDict._removeDupes()` is
    # not run. Otherwise bond type 12 and 21 will be seen as duplicates
    # and combined.
    def test_bond_no_reversal(self):
        universe = mda.Universe(LAMMPSdata_many_bonds, format="DATA")
        nbonds = 22
        bondtypes = universe.atoms.bonds.types()
        assert len(bondtypes) == nbonds

    def test_angles_types(self, PSFDCD):
        """TopologyDict for angles"""
        assert len(PSFDCD.atoms.angles.types()) == 130

    def test_angles_contains(self, a_td):
        assert ("23", "73", "1") in a_td

    def test_angles_uniqueness(self, a_td):
        bondtypes = a_td.keys()
        assert not any(b[::-1] in bondtypes for b in bondtypes if b[::-1] != b)

    def test_angles_reversal(self, a_td):
        bondtypes = list(a_td.keys())
        b = bondtypes[1]
        assert all([b in a_td, b[::-1] in a_td])

        tg1 = a_td[b]
        tg2 = a_td[b[::-1]]
        assert tg1 == tg2

    def test_dihedrals_types(self, PSFDCD):
        """TopologyDict for dihedrals"""
        assert len(PSFDCD.atoms.dihedrals.types()) == 220

    def test_dihedrals_contains(self, t_td):
        assert ("30", "29", "20", "70") in t_td

    def test_dihedrals_uniqueness(self, t_td):
        bondtypes = t_td.keys()
        assert not any(b[::-1] in bondtypes for b in bondtypes if b[::-1] != b)

    def test_dihedrals_reversal(self, t_td):
        bondtypes = list(t_td.keys())
        b = bondtypes[1]
        assert all([b in t_td, b[::-1] in t_td])

        tg1 = t_td[b]
        tg2 = t_td[b[::-1]]
        assert tg1 == tg2

    def test_bad_creation(self):
        """Test making a TopologyDict out of nonsense"""
        inputlist = ["a", "b", "c"]
        with pytest.raises(TypeError):
            TopologyDict(inputlist)

    def test_bad_creation_TG(self):
        """Test making a TopologyGroup out of nonsense"""
        inputlist = ["a", "b", "c"]
        with pytest.raises(TypeError):
            TopologyGroup(inputlist)

    def test_tg_creation_bad_btype(self, PSFDCD):
        vals = np.array([[0, 10], [5, 15]])
        with pytest.raises(ValueError):
            TopologyGroup(vals, PSFDCD, btype="apple")

    def test_bond_tg_creation_notype(self, PSFDCD):
        vals = np.array([[0, 10], [5, 15]])
        tg = TopologyGroup(vals, PSFDCD)

        assert tg.btype == "bond"
        assert_equal(tg[0].indices, (0, 10))
        assert_equal(tg[1].indices, (5, 15))

    def test_angle_tg_creation_notype(self, PSFDCD):
        vals = np.array([[0, 5, 10], [5, 10, 15]])
        tg = TopologyGroup(vals, PSFDCD)

        assert tg.btype == "angle"
        assert_equal(tg[0].indices, (0, 5, 10))
        assert_equal(tg[1].indices, (5, 10, 15))

    def test_dihedral_tg_creation_notype(self, PSFDCD):
        vals = np.array([[0, 2, 4, 6], [5, 7, 9, 11]])

        tg = TopologyGroup(vals, PSFDCD)

        assert tg.btype == "dihedral"
        assert_equal(tg[0].indices, (0, 2, 4, 6))
        assert_equal(tg[1].indices, (5, 7, 9, 11))

    def test_create_guessed_tg(self, PSFDCD):
        vals = np.array([[0, 10], [5, 15]])
        tg = TopologyGroup(vals, PSFDCD, guessed=True)

        assert_equal(tg._guessed, np.array([True, True]))

    def test_create_guessed_tg_2(self, PSFDCD):
        vals = np.array([[0, 10], [5, 15]])
        tg = TopologyGroup(vals, PSFDCD, guessed=False)

        assert_equal(tg._guessed, np.array([False, False]))

    def test_TG_equality(self, PSFDCD):
        """Make two identical TGs,
        * check they're equal
        * change one very slightly and see if they notice
        """
        tg = PSFDCD.atoms.bonds.selectBonds(("23", "3"))
        tg2 = PSFDCD.atoms.bonds.selectBonds(("23", "3"))

        assert tg == tg2

        tg3 = PSFDCD.atoms.bonds.selectBonds(("81", "10"))

        assert not (tg == tg3)
        assert tg != tg3

    def test_create_TopologyGroup(self, res1, PSFDCD):
        res1_tg = res1.atoms.bonds.select_bonds(("23", "3"))  # make a tg
        assert len(res1_tg) == 4  # check size of tg
        testbond = PSFDCD.atoms[7].bonds[0]
        assert testbond in res1_tg  # check a known bond is present

        res1_tg2 = res1.atoms.bonds.select_bonds(("23", "3"))
        assert res1_tg == res1_tg2

    @pytest.mark.parametrize(
        "attr", ["bonds", "angles", "dihedrals", "impropers"]
    )
    def test_TG_loose_intersection(self, PSFDCD, attr):
        """Pull bonds from a TG which are at least partially in an AG"""
        ag = PSFDCD.atoms[10:60]

        TG = getattr(PSFDCD.atoms, attr)
        ref = getattr(ag, attr)
        newTG = TG.atomgroup_intersection(ag)

        assert newTG == ref

    def test_TG_strict_intersection(self, PSFDCD):
        """Pull bonds from TG which are fully in an AG"""

        def check_strict_intersection(topg, atomg):
            new_topg = topg.atomgroup_intersection(atomg, strict=True)

            return all([all([a in atomg for a in b.atoms]) for b in new_topg])

        def manual(topg, atomg):
            """Gives a set of the Bonds that should be found"""
            if len(atomg) == 1:  # hack for Atom input
                atomg = [atomg]
            man = []
            for b in topg:
                if all([a in atomg for a in b.atoms]):
                    man.append(b)

            return set(man)

        testinput = PSFDCD.atoms[10:60]

        # bonds
        assert check_strict_intersection(PSFDCD.atoms.bonds, testinput)
        assert manual(PSFDCD.atoms.bonds, testinput) == set(
            PSFDCD.atoms.bonds.atomgroup_intersection(testinput, strict=True)
        )
        # angles
        assert check_strict_intersection(PSFDCD.atoms.angles, testinput)
        assert manual(PSFDCD.atoms.angles, testinput) == set(
            PSFDCD.atoms.angles.atomgroup_intersection(testinput, strict=True)
        )
        # dihedrals
        assert check_strict_intersection(PSFDCD.atoms.dihedrals, testinput)
        assert manual(PSFDCD.atoms.dihedrals, testinput) == set(
            PSFDCD.atoms.dihedrals.atomgroup_intersection(
                testinput, strict=True
            )
        )

    def test_add_TopologyGroups(self, res1, res2, PSFDCD):
        res1_tg = res1.atoms.bonds.selectBonds(("23", "3"))
        res2_tg = res2.atoms.bonds.selectBonds(("23", "3"))

        combined_tg = res1_tg + res2_tg  # add tgs together
        assert len(combined_tg) == 10

        big_tg = PSFDCD.atoms.bonds.selectBonds(("23", "3"))
        big_tg += combined_tg  # try and add some already included bonds
        assert len(big_tg) == 494  # check len doesn't change

    def test_add_empty_to_TG(self, PSFDCD):
        tg1 = PSFDCD.bonds[10:15]
        tg2 = PSFDCD.bonds[:0]  # empty
        tg3 = tg1 + tg2

        assert tg1 == tg3

    def test_add_TO_to_empty_TG(self, PSFDCD):
        tg1 = PSFDCD.bonds[:0]  # empty
        to = PSFDCD.bonds[5]
        tg3 = tg1 + to

        assert_equal(tg3.indices, to.indices[None, :])

    def test_add_TG_to_empty_TG(self, PSFDCD):
        tg1 = PSFDCD.bonds[:0]  # empty
        tg2 = PSFDCD.bonds[5:7]
        tg3 = tg1 + tg2

        assert tg2 == tg3

    def test_add_singleitem(self, PSFDCD):
        tg = PSFDCD.atoms.bonds[:10]
        to = PSFDCD.atoms.bonds[55]

        assert len(tg + to) == 11

    def test_add_wrongtype_TopologyGroup(self, PSFDCD):
        tg = PSFDCD.atoms.bonds[:10]  # TG of bonds
        ang = PSFDCD.atoms.angles[10]  # single angle
        angg = PSFDCD.atoms.angles[:10]  # TG of angles

        for other in [ang, angg]:
            with pytest.raises(TypeError):
                this = tg + other

    def test_bad_add_TopologyGroup(self, PSFDCD):
        tg = PSFDCD.atoms.bonds[:10]  # TopologyGroup
        ag = PSFDCD.atoms[:10]  # AtomGroup
        with pytest.raises(TypeError):
            this = tg + ag

    def test_TG_getitem_single(self, PSFDCD):
        tg = PSFDCD.atoms.bonds[:10]
        bondlist = list(tg)
        bond = tg[0]

        assert bond == bondlist[0]

    def test_TG_getitem_slice(self, PSFDCD):
        tg = PSFDCD.atoms.bonds[:10]
        tg2 = tg[1:4]

        assert_equal(tg2.indices, tg.indices[1:4])

    def test_TG_getitem_fancy(self, PSFDCD):
        tg = PSFDCD.atoms.bonds[:10]
        tg2 = tg[[1, 4, 5]]

        manual = TopologyGroup(tg.indices[[1, 4, 5]], tg.universe, tg.btype)

        assert list(tg2) == list(manual)

    def test_TG_getitem_bool(self, PSFDCD):
        # Issue #282
        sel = np.array([True, False, True])
        tg = PSFDCD.atoms.bonds[:3]
        tg2 = tg[sel]

        assert len(tg2) == 2
        for b in [tg[0], tg[2]]:
            assert b in tg2

    def test_TG_getitem_bool_IE(self, PSFDCD):
        sel = []
        tg = PSFDCD.atoms.bonds[10:13]
        tg2 = tg[sel]

        assert len(tg2) == 0

    # atomX access
    def test_atom1(self, PSFDCD):
        tg = PSFDCD.bonds[:5]
        a1 = tg.atom1

        assert len(tg) == len(a1)
        for atom, bond in zip(a1, tg):
            assert atom == bond[0]

    def test_atom2(self, PSFDCD):
        tg = PSFDCD.bonds[:5]
        a2 = tg.atom2

        assert len(tg) == len(a2)
        for atom, bond in zip(a2, tg):
            assert atom == bond[1]

    def test_atom3_IE(self, PSFDCD):
        tg = PSFDCD.bonds[:5]
        with pytest.raises(IndexError):
            tg.atom3

    def test_atom3(self, PSFDCD):
        tg = PSFDCD.angles[:5]
        a3 = tg.atom3
        assert len(tg) == len(a3)
        for atom, bond in zip(a3, tg):
            assert atom == bond[2]

    def test_atom4_IE(self, PSFDCD):
        tg = PSFDCD.bonds[:5]
        with pytest.raises(IndexError):
            tg.atom4

    def test_atom4(self, PSFDCD):
        tg = PSFDCD.dihedrals[:5]

        a4 = tg.atom4
        assert len(tg) == len(a4)
        for atom, bond in zip(a4, tg):
            assert atom == bond[3]


class TestTopologyGroup_Cython(object):
    """
    Check that the shortcut to all cython functions:
     - work (return proper values)
     - catch errors
    """

    @staticmethod
    @pytest.fixture
    def bgroup(PSFDCD):
        return PSFDCD.atoms[:5].bonds

    @staticmethod
    @pytest.fixture
    def agroup(PSFDCD):
        return PSFDCD.atoms[:5].angles

    @staticmethod
    @pytest.fixture
    def dgroup(PSFDCD):
        return PSFDCD.atoms[:5].dihedrals

    @staticmethod
    @pytest.fixture
    def igroup(PSFDCD):
        return PSFDCD.atoms[:5].impropers

    # bonds
    def test_wrong_type_bonds(self, agroup, dgroup, igroup):
        for tg in [agroup, dgroup, igroup]:
            with pytest.raises(TypeError):
                tg.bonds()

    def test_right_type_bonds(self, bgroup, PSFDCD):
        assert_equal(
            bgroup.bonds(),
            calc_bonds(bgroup.atom1.positions, bgroup.atom2.positions),
        )
        assert_equal(
            bgroup.bonds(pbc=True),
            calc_bonds(
                bgroup.atom1.positions,
                bgroup.atom2.positions,
                box=PSFDCD.dimensions,
            ),
        )
        assert_equal(
            bgroup.values(),
            calc_bonds(bgroup.atom1.positions, bgroup.atom2.positions),
        )
        assert_equal(
            bgroup.values(pbc=True),
            calc_bonds(
                bgroup.atom1.positions,
                bgroup.atom2.positions,
                box=PSFDCD.dimensions,
            ),
        )

    # angles
    def test_wrong_type_angles(self, bgroup, dgroup, igroup):
        for tg in [bgroup, dgroup, igroup]:
            with pytest.raises(TypeError):
                tg.angles()

    def test_right_type_angles(self, agroup, PSFDCD):
        assert_equal(
            agroup.angles(),
            calc_angles(
                agroup.atom1.positions,
                agroup.atom2.positions,
                agroup.atom3.positions,
            ),
        )
        assert_equal(
            agroup.angles(pbc=True),
            calc_angles(
                agroup.atom1.positions,
                agroup.atom2.positions,
                agroup.atom3.positions,
                box=PSFDCD.dimensions,
            ),
        )
        assert_equal(
            agroup.values(),
            calc_angles(
                agroup.atom1.positions,
                agroup.atom2.positions,
                agroup.atom3.positions,
            ),
        )
        assert_equal(
            agroup.values(pbc=True),
            calc_angles(
                agroup.atom1.positions,
                agroup.atom2.positions,
                agroup.atom3.positions,
                box=PSFDCD.dimensions,
            ),
        )

    # dihedrals & impropers
    def test_wrong_type_dihedrals(self, bgroup, agroup):
        for tg in [bgroup, agroup]:
            with pytest.raises(TypeError):
                tg.dihedrals()

    def test_right_type_dihedrals(self, dgroup, PSFDCD):
        assert_equal(
            dgroup.dihedrals(),
            calc_dihedrals(
                dgroup.atom1.positions,
                dgroup.atom2.positions,
                dgroup.atom3.positions,
                dgroup.atom4.positions,
            ),
        )
        assert_equal(
            dgroup.dihedrals(pbc=True),
            calc_dihedrals(
                dgroup.atom1.positions,
                dgroup.atom2.positions,
                dgroup.atom3.positions,
                dgroup.atom4.positions,
                box=PSFDCD.dimensions,
            ),
        )
        assert_equal(
            dgroup.values(),
            calc_dihedrals(
                dgroup.atom1.positions,
                dgroup.atom2.positions,
                dgroup.atom3.positions,
                dgroup.atom4.positions,
            ),
        )
        assert_equal(
            dgroup.values(pbc=True),
            calc_dihedrals(
                dgroup.atom1.positions,
                dgroup.atom2.positions,
                dgroup.atom3.positions,
                dgroup.atom4.positions,
                box=PSFDCD.dimensions,
            ),
        )

    def test_right_type_impropers(self, igroup, PSFDCD):
        assert_equal(
            igroup.dihedrals(),
            calc_dihedrals(
                igroup.atom1.positions,
                igroup.atom2.positions,
                igroup.atom3.positions,
                igroup.atom4.positions,
            ),
        )
        assert_equal(
            igroup.dihedrals(pbc=True),
            calc_dihedrals(
                igroup.atom1.positions,
                igroup.atom2.positions,
                igroup.atom3.positions,
                igroup.atom4.positions,
                box=PSFDCD.dimensions,
            ),
        )
        assert_equal(
            igroup.values(),
            calc_dihedrals(
                igroup.atom1.positions,
                igroup.atom2.positions,
                igroup.atom3.positions,
                igroup.atom4.positions,
            ),
        )
        assert_equal(
            igroup.values(pbc=True),
            calc_dihedrals(
                igroup.atom1.positions,
                igroup.atom2.positions,
                igroup.atom3.positions,
                igroup.atom4.positions,
                box=PSFDCD.dimensions,
            ),
        )


def test_bond_length_pbc():
    u = mda.Universe(TRZ_psf, TRZ)

    ref = u.bonds[0].length()

    # move an atom a box width in all dimensions
    u.atoms[0].position += u.dimensions[:3]

    assert_almost_equal(ref, u.bonds[0].length(pbc=True), decimal=6)


def test_cross_universe_eq():
    u1 = mda.Universe(PSF)
    u2 = mda.Universe(PSF)

    assert not (u1.bonds[0] == u2.bonds[0])


def test_zero_size_TG_indices_bonds():
    u = mda.Universe.empty(10)

    u.add_TopologyAttr("bonds", values=[(1, 2), (2, 3)])

    ag = u.atoms[[0]]

    idx = ag.bonds.to_indices()

    assert idx.shape == (0, 2)
    assert idx.dtype == np.int32


def test_zero_size_TG_indices_angles():
    u = mda.Universe.empty(10)

    u.add_TopologyAttr("angles", values=[(1, 2, 3), (2, 3, 4)])

    ag = u.atoms[[0]]

    idx = ag.angles.to_indices()

    assert idx.shape == (0, 3)
    assert idx.dtype == np.int32



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""Stub module for distopia --- :mod:`MDAnalysis.analysis.distopia`
===================================================================

This module is a stub to provide distopia distance functions to `distances.py`
as a selectable backend.
"""
import warnings
from packaging.version import Version

MIN_DISTOPIA_VERSION = Version("0.4.0")

# check for distopia
try:
    import distopia
except ImportError:
    HAS_DISTOPIA = False
    distopia_version = Version("0.0.0")
else:
    HAS_DISTOPIA = True

    # check for compatibility: currently needs to be >=0.4.0,
    # some versions of `distopia` don't have a version attribute
    try:
        distopia_version = Version(distopia.__version__)
    except AttributeError:
        distopia_version = Version("0.0.0")
    if distopia_version < MIN_DISTOPIA_VERSION:
        warnings.warn(
            f"distopia version {distopia_version} is too old; "
            f"need at least {MIN_DISTOPIA_VERSION}, Your installed version of "
            "distopia will NOT be used.",
            category=RuntimeWarning,
        )
        HAS_DISTOPIA = False


import numpy as np


def calc_bond_distance_ortho(
    coords1: np.ndarray,
    coords2: np.ndarray,
    box: np.ndarray,
    results: np.ndarray,
) -> None:
    distopia.distances_ortho(coords1, coords2, box[:3], results=results)


def calc_bond_distance(
    coords1: np.ndarray, coords2: np.ndarray, results: np.ndarray
) -> None:
    distopia.distances_no_box(coords1, coords2, results=results)


def calc_bond_distance_triclinic(
    coords1: np.ndarray,
    coords2: np.ndarray,
    box: np.ndarray,
    results: np.ndarray,
) -> None:
    distopia.distances_triclinic(coords1, coords2, box, results=results)


def calc_angle(
    coords1: np.ndarray,
    coords2: np.ndarray,
    coords3: np.ndarray,
    results: np.ndarray,
) -> None:
    distopia.angles_no_box(coords1, coords2, coords3, results=results)


def calc_angle_ortho(
    coords1: np.ndarray,
    coords2: np.ndarray,
    coords3: np.ndarray,
    box: np.ndarray,
    results: np.ndarray,
) -> None:
    distopia.angles_ortho(coords1, coords2, coords3, box[:3], results=results)


def calc_angle_triclinic(
    coords1: np.ndarray,
    coords2: np.ndarray,
    coords3: np.ndarray,
    box: np.ndarray,
    results: np.ndarray,
) -> None:

    distopia.angles_triclinic(coords1, coords2, coords3, box, results=results)


def calc_dihedral(
    coords1: np.ndarray,
    coords2: np.ndarray,
    coords3: np.ndarray,
    coords4: np.ndarray,
    results: np.ndarray,
) -> None:
    distopia.dihedrals_no_box(
        coords1, coords2, coords3, coords4, results=results
    )


def calc_dihedral_ortho(
    coords1: np.ndarray,
    coords2: np.ndarray,
    coords3: np.ndarray,
    coords4: np.ndarray,
    box: np.ndarray,
    results: np.ndarray,
) -> None:
    distopia.dihedrals_ortho(
        coords1, coords2, coords3, coords4, box[:3], results=results
    )


def calc_dihedral_triclinic(
    coords1: np.ndarray,
    coords2: np.ndarray,
    coords3: np.ndarray,
    coords4: np.ndarray,
    box: np.ndarray,
    results: np.ndarray,
) -> None:
    distopia.dihedrals_triclinic(
        coords1, coords2, coords3, coords4, box, results=results
    )


def calc_distance_array(
    coords1: np.ndarray, coords2: np.ndarray, results: np.ndarray
) -> None:
    distopia.distance_array_no_box(coords1, coords2, results=results)


def calc_distance_array_ortho(
    coords1: np.ndarray,
    coords2: np.ndarray,
    box: np.ndarray,
    results: np.ndarray,
) -> None:
    distopia.distance_array_ortho(coords1, coords2, box[:3], results=results)


def calc_distance_array_triclinic(
    coords1: np.ndarray,
    coords2: np.ndarray,
    box: np.ndarray,
    results: np.ndarray,
) -> None:
    distopia.distance_array_triclinic(coords1, coords2, box, results=results)


def calc_self_distance_array(coords: np.ndarray, results: np.ndarray) -> None:
    distopia.self_distance_array_no_box(coords, results=results)


def calc_self_distance_array_ortho(
    coords: np.ndarray, box: np.ndarray, results: np.ndarray
) -> None:
    distopia.self_distance_array_ortho(coords, box[:3], results=results)


def calc_self_distance_array_triclinic(
    coords: np.ndarray, box: np.ndarray, results: np.ndarray
) -> None:
    distopia.self_distance_array_triclinic(coords, box, results=results)
